from sklearn.base import BaseEstimator, TransformerMixin
from pandas.core.dtypes.common import is_numeric_dtype
import logging

from sklearn.neighbors import KNeighborsRegressor,KNeighborsClassifier
from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier
import xgboost as xgb 
import lightgbm as lgb
import pandas as pd


class Model_Imputer:
    """
    Parameters
    ----------
    classifier : estimator object.
    A object of that type is instantiated for each search point.
    This object is assumed to implement the scikit-learn estimator api.
    regressor : estimator object.
    A object of that type is instantiated for each search point.
    This object is assumed to implement the scikit-learn estimator api.
     n_iter : int
     Determines the number of iteration.
     initial_guess : string, callable or None, default='median'
     If ``mean``, the initial impuatation will use the median of the features.
     If ``median``, the initial impuatation will use the median of the features.
    """

    def __init__(self, method : str='MISSFOREST', initial_guess: str='median', n_iter: int=5):
        
        self.models = {
                "MISSFOREST" : [RandomForestClassifier(), RandomForestRegressor()],
                "KNN" : [KNeighborsClassifier(n_neighbors=3),KNeighborsRegressor(n_neighbors=3)], 
                "XGB" : [xgb.XGBClassifier(),xgb.XGBRegressor()], 
                "LGBM" : [lgb.LGBMClassifier(),lgb.LGBMRegressor()],  
            }
        self.classifier = self.models[method][0]
        self.regressor = self.models[method][1]
        self.initial_guess = initial_guess
        self.n_iter = n_iter

    def fit_transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
        Dataset (features only) that needed to be imputed.
        """
        miss_row = {}
        for c in X.columns:
            miss_row[c] = X[c][X[c].isnull() > 0].index
        miss_col = [k for k in miss_row.keys() if len(miss_row[k]) > 0]
        obs_row = X[X.isnull().sum(axis=1) == 0].index

        mappings = {}
        rev_mappings = {}
        for c in X.columns:
            if type(X[c].dropna().sample(n=1).values[0]) == str:
                mappings[c] = {k: v for k, v in zip(X[c].dropna().unique(), range(X[c].dropna().nunique()))}
                rev_mappings[c] = {v: k for k, v in zip(X[c].dropna().unique(), range(X[c].dropna().nunique()))}

        non_impute_cols = [c for c in X.columns if c not in mappings.keys()]
        # 1) Make an initial guess for all missing categorical/numeric values (e.g. mean, mode)
        for c in X.columns:
            # if datatype is numeric, fillna with mean or median
            if X[c].dtype in ['int_', 'int8', 'int16', 'int32', 'int64', 'uint8', 'uint16', 'uint32', 'uint64',
                              'float_','float16', 'float32', 'float64']:

                if self.initial_guess == 'mean':
                    X[c].fillna(X[c].mean(), inplace=True)

                else:
                    X[c].fillna(X[c].median(), inplace=True)

            # if datatype is categorical, fillna with mode
            else:
                X[c].fillna(X[c].mode().values[0], inplace=True)

        # Label Encoding
        for c in mappings:
            X[c].replace(mappings[c], inplace=True)
            X[c] = X[c].astype(int)

        iter = 0
        while True:
            for c in miss_col:
                if c in mappings:
                    estimator = self.classifier
                else:
                    estimator = self.regressor

                # Fit estimator with imputed X
                estimator.fit(X.drop(c, axis=1).loc[obs_row], X[c].loc[obs_row])

                # Predict the missing column with the trained estimator
                y_pred = estimator.predict(X.loc[miss_row[c]].drop(c, axis=1))
                y_pred = pd.Series(y_pred)
                y_pred.index = miss_row[c]
                # Update imputed matrix
                X.loc[miss_row[c], c] = y_pred
            # Check if Criteria is met
            if iter >= self.n_iter:
                break

            iter += 1
        # Reverse mapping
        for c in rev_mappings:
            X[c].replace(rev_mappings[c], inplace=True)

        self.X = X

        return X

##########################################################################
# column 순서대로 제공되는 impute_strategies에 따라 결측값을 대체합니다.
# fit에서 train set으로부터 결측값을 대체할 imputing_values를 추출하고,
# transform에서 strategy에 해당하는 방법에 따라 결측값을 처리합니다. 
##########################################################################
class AccutuningNullImputerBycol(BaseEstimator, TransformerMixin):
    def __init__(self, columns_name, impute_strategies):
        self.columns_name = columns_name
        self.impute_strategies = impute_strategies

    def fit(self, X, y=0, **fit_params):
        self.strategies_dict = dict(zip(self.columns_name, self.impute_strategies))
        imputing_values = []
        for col in self.strategies_dict:
            try:
                imputing_col = X.loc[:, col]
            except KeyError:
                logging.critical(
                    'No such column name in the dataset - Null Imputer'
                )
            else:
                strategy = self.strategies_dict[col]
                val = None
                if strategy == 'MOST_FREQUENT':
                    val = imputing_col.mode()[0]
                elif strategy == 'UNKNOWN':
                    val = 'Unknown'
                elif strategy in ('MISSFOREST','KNN','XGB','LGBM'):
                    val = 'tbd'
                elif is_numeric_dtype(imputing_col):
                    if strategy == 'MEAN':
                        val = imputing_col.mean()
                    elif strategy == 'MEDIAN':
                        val = imputing_col.median()
                    elif strategy == 'ZERO':
                        val = 0
                    elif strategy == 'MINIMUM':
                        val = imputing_col.min()
                imputing_values.append(val)
        
        self.imputing_dict = dict(zip(self.columns_name, imputing_values))
        return self

    def transform(self, X, y=0):
        X_tr = X.copy()

        if 'MISSFOREST' in self.strategies_dict.values() :
            imputer = Model_Imputer()
            MISSFOREST_imputed = imputer.fit_transform(X_tr)
        if 'KNN' in self.strategies_dict.values() :
            imputer = Model_Imputer(method = 'KNN')
            KNN_imputed = imputer.fit_transform(X_tr)
        if 'XGB' in self.strategies_dict.values() :
            imputer = Model_Imputer(method = 'XGB')
            XGB_imputed = imputer.fit_transform(X_tr)
        if 'LGBM' in self.strategies_dict.values() :
            imputer = Model_Imputer(method = 'LGBM')
            LGBM_imputed = imputer.fit_transform(X_tr)


        # impute_strategies는 column 순서대로 그에 해당하는 impute strategy list
        for col in self.strategies_dict:
            imputing_val = self.imputing_dict[col]
            strategy = self.strategies_dict[col]
            # numerical, categorical 공통 impute 방법
            if strategy in ('NONE', '', None):
                pass
            elif strategy == 'DROP':
                X_tr = X_tr.dropna(how='any', subset=[col], axis=0)
            # categorical impute 방법
            elif strategy == 'UNKNOWN':
                X_tr.loc[:, col] = X_tr.loc[:, col].fillna(value='Unknown')
            # numerical impute 방법
            elif strategy in ('MOST_FREQUENT', 'Unknown', 'MEAN', 'MEDIAN', 'ZERO', 'MINIMUM') and imputing_val is not None:
                X_tr.loc[:, col] = X_tr.loc[:, col].fillna(imputing_val)
            # model 기반 impute : categorical, numerical 둘 다 가능
            elif strategy == 'MISSFOREST' :
                X_tr.loc[:, col] = MISSFOREST_imputed.loc[:, col]
            elif strategy == 'KNN' :
                X_tr.loc[:, col] = KNN_imputed.loc[:, col]
            elif strategy == 'XGB' :
                X_tr.loc[:, col] = XGB_imputed.loc[:, col]
            elif strategy == 'LGBM' :
                X_tr.loc[:, col] = LGBM_imputed.loc[:, col]

        return X_tr



from sklearn.base import BaseEstimator, TransformerMixin
import pandas as pd
import numpy as np

##########################################################################
# 선택된 Column에 대하여 Lag Columns를 생성하여 추가합니다.
# lag만 제공될 경우 해당하는 단일 lag column을 생성하고, 
# lag, lag2 모두 제공될 경우 lag ~ lag2 범위에 해당하는 lag columns를 생성합니다.
# lag 값이 양수일 경우에는 컬럼 값을 아래로 내려 과거를 나타내고, 음수일 경우에는 위로 올려 미래를 나타냅니다. (pandas shift 참조)
##########################################################################
class AccutuningLagColumnAdder(BaseEstimator, TransformerMixin):
    def __init__(self, target_cols=None, lag=None, lag2=None):
        self.target_cols = target_cols
        self.lag = lag
        self.lag2 = lag2

    def fit(self, X, y=0, **fit_params):
        self.row_size = X.shape[0]
        return self

    def transform(self, X, y=0, **fit_params):
        X_tr = X.copy()
        for target_col in self.target_cols:
            if self.lag2 is not None:
                lag_range = [
                    int(x) for x in range(self.lag, self.lag2 + 1)
                    if x != 0
                ]
                for i in lag_range:
                    col_name = '{}_lag{}'.format(target_col, i)

                    if target_col in X.columns:
                        if i < self.row_size:
                            tmp_col = pd.Series(X[target_col].shift(np.negative(i)))
                        else:
                            tmp_col = pd.Series(X[target_col]) # shift가 안 될 경우 그냥 해당 컬럼 추가 (predict할 때 컬럼 수 맞추기 위하여)

                        filler = 'ffill' if i > 0 else 'bfill'
                        tmp_col = tmp_col.fillna(method=filler)
                        X_tr.insert(
                            X.columns.get_loc(target_col),
                            col_name,
                            tmp_col
                        )
                    else:
                        X_tr[col_name] = np.NaN
            else:
                col_name = '{}_lag{}'.format(target_col, self.lag)
                if target_col in X.columns:
                    if self.lag < self.row_size:
                        tmp_col = pd.Series(X[target_col].shift(np.negative(self.lag)))
                    else:
                        tmp_col = pd.Series(X[target_col]) # shift가 안 될 경우 그냥 해당 컬럼 추가 (predict할 때 컬럼 수 맞추기 위하여)
                    filler = 'ffill' if self.lag > 0 else 'bfill'
                    tmp_col = tmp_col.fillna(method=filler)
                    X_tr.insert(
                        X.columns.get_loc(target_col),
                        col_name,
                        tmp_col
                    )
                else:
                    X_tr[col_name] = np.NaN
        return X_tr



from sklearn.base import BaseEstimator, TransformerMixin

##########################################################################
# column_names에 해당하는 Columns만 필터링합니다.
##########################################################################
class AccutuningColumnSelector(BaseEstimator, TransformerMixin):
    def __init__(self, column_names):
        self.column_names = column_names

    def fit(self, X, y=0, **fit_params):
        return self

    def transform(self, X, y=0):
        inter = set(X.columns) & set(self.column_names)
        filtered = [
            c
            for c in self.column_names
            if c in inter
        ]  # 순서보존을 위해 inter사용대신 재작업
        return X.loc[:, filtered]



import pandas as pd
# import numpy as np
from datetime import datetime
from sklearn.base import BaseEstimator, TransformerMixin
import dateutil.parser


def datetime_intervals(X, datetime_col):
    X.loc[:, datetime_col] = pd.to_datetime(X.loc[:, datetime_col], errors='coerce')
    X = X.sort_values(by=datetime_col)
    intervals = (X[datetime_col] - X[datetime_col].shift(1)).dropna()
    if intervals.nunique() == 1:
        fixed_interval = True
    else:
        fixed_interval = False
    return intervals, fixed_interval


# datetime이 있는 데이터를 일정한 간격으로 resample 합니다.
class AccutuningTimeseriesResample(BaseEstimator, TransformerMixin):
    def __init__(self, datetime_col, datetime_format=None, interval=None):
        self.datetime_col = datetime_col
        self.datetime_format = datetime_format
        self.interval = interval

    def fit(self, X, y=0, **fit_params):
        self.intervals, self.fixed_interval = datetime_intervals(X, self.datetime_col)
        return self

    def transform(self, X, y=0, **fit_params):
        target_column = X.loc[:, self.datetime_col]
        if self.datetime_format:
            X.loc[:, self.datetime_col] = pd.to_datetime(
                target_column,
                format=self.datetime_format,
                errors='coerce'
            )
        else:
            try:
                X.loc[:, self.datetime_col] = pd.to_datetime(
                    target_column,
                    infer_datetime_format=True
                )
            except Exception:
                def _parse(x):
                    try:
                        # TODO: 매번 format을 유추하기 때문에 느리다;
                        ret = dateutil.parser.parse(x)
                    except Exception:
                        ret = datetime.fromtimestamp(0)
                    return ret
                X.loc[:, self.datetime_col] = target_column.map(lambda x: _parse(x))
        X_tr = X.copy()
        X_tr.index = X_tr.loc[:, self.datetime_col]
        X_tr = X_tr.sort_index()
        if self.fixed_interval:
            pass
        elif self.interval is None:
            self.interval = self.intervals.mode()[0]
            X_tr = X_tr.resample(self.interval).mean()
        else:
            X_tr = X_tr.resample(self.interval).mean()
        return X_tr


# 일정한 간격으로 timeseries 데이터를 재배치하면서 결측값이 생기는 경우 Interpolate 합니다.
class AccutuningTimeseriesInterpolate(BaseEstimator, TransformerMixin):
    def __init__(self, method='linear'):
        self.method = method

    def fit(self, X, y=0, **fit_params):
        if X.isnull().values.any():
            self.need_interpolation = True
        else:
            self.need_interpolation = False
        return self

    def transform(self, X, y=0, **fit_params):
        if self.need_interpolation:
            X_tr = X.interpolate(method=self.method)
        else:
            X_tr = X.copy()
        return X_tr


# 지정한 Window size와 Moving Average 방식대로 이동평균 컬럼을 추가합니다.
class AccutuningMovingAverage(BaseEstimator, TransformerMixin):
    """
    [(colname1, 5, 'ma'), (colname1, 10, 'ema'), (colname2, 5, 'ema')]
    """
    def __init__(self, how):
        self.how = how

    def fit(self, X, y=0, **fit_params):
        self.newcols_dict = {}

        for i in self.how:
            col, window, method = i[0], i[1], i[2]

            try:
                if method == 'ma':
                    newcol = X[col].rolling(window).mean()
                elif method == 'ema':
                    newcol = X[col].ewm(window).mean()

                colname = str(col) + '_' + str(window) + '_' + str(method)
                newcol.name = colname

                if col in self.newcols_dict.keys():
                    self.newcols_dict[col].append(newcol)
                else:
                    self.newcols_dict[col] = []
                    self.newcols_dict[col].append(newcol)

            except Exception as e:
                pass

        return self

    def transform(self, X, y=0, **fit_params):
        X_tr = X.copy()
        for i in self.newcols_dict.keys():
            target_idx = X.columns.get_loc(i)
            X_front = X_tr.columns[:target_idx + 1]
            X_back = X_tr.columns[target_idx + 1:]
            converted = pd.concat(self.newcols_dict[i], axis=1)
            X_tr = pd.concat([X_tr[X_front], converted, X_tr[X_back]], axis=1)
        X_tr = X_tr.dropna()
        return X_tr



# Temporarily use sklearn.preprocessing.OrdinalEncoder from sklearn==0.24.0

from sklearn.base import BaseEstimator, TransformerMixin

import numbers
from sklearn.utils import check_array
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.validation import _deprecate_positional_args

from typing import NamedTuple

import numpy as np

def is_scalar_nan(x):
    """Tests if x is NaN.
    This function is meant to overcome the issue that np.isnan does not allow
    non-numerical types as input, and that np.nan is not float('nan').
    Parameters
    ----------
    x : any type
    Returns
    -------
    boolean
    Examples
    --------
    >>> is_scalar_nan(np.nan)
    True
    >>> is_scalar_nan(float("nan"))
    True
    >>> is_scalar_nan(None)
    False
    >>> is_scalar_nan("")
    False
    >>> is_scalar_nan([np.nan])
    False
    """
    # convert from numpy.bool_ to python bool to ensure that testing
    # is_scalar_nan(x) is True does not fail.
    return bool(isinstance(x, numbers.Real) and np.isnan(x))


def _unique(values, *, return_inverse=False):
    """Helper function to find unique values with support for python objects.
    Uses pure python method for object dtype, and numpy method for
    all other dtypes.
    Parameters
    ----------
    values : ndarray
        Values to check for unknowns.
    return_inverse : bool, default=False
        If True, also return the indices of the unique values.
    Returns
    -------
    unique : ndarray
        The sorted unique values.
    unique_inverse : ndarray
        The indices to reconstruct the original array from the unique array.
        Only provided if `return_inverse` is True.
    """
    if values.dtype == object:
        return _unique_python(values, return_inverse=return_inverse)
    # numerical
    out = np.unique(values, return_inverse=return_inverse)

    if return_inverse:
        uniques, inverse = out
    else:
        uniques = out

    # np.unique will have duplicate missing values at the end of `uniques`
    # here we clip the nans and remove it from uniques
    if uniques.size and is_scalar_nan(uniques[-1]):
        nan_idx = np.searchsorted(uniques, np.nan)
        uniques = uniques[:nan_idx + 1]
        if return_inverse:
            inverse[inverse > nan_idx] = nan_idx

    if return_inverse:
        return uniques, inverse
    return uniques


class MissingValues(NamedTuple):
    """Data class for missing data information"""
    nan: bool
    none: bool

    def to_list(self):
        """Convert tuple to a list where None is always first."""
        output = []
        if self.none:
            output.append(None)
        if self.nan:
            output.append(np.nan)
        return output


def _extract_missing(values):
    """Extract missing values from `values`.
    Parameters
    ----------
    values: set
        Set of values to extract missing from.
    Returns
    -------
    output: set
        Set with missing values extracted.
    missing_values: MissingValues
        Object with missing value information.
    """
    missing_values_set = {value for value in values
                          if value is None or is_scalar_nan(value)}

    if not missing_values_set:
        return values, MissingValues(nan=False, none=False)

    if None in missing_values_set:
        if len(missing_values_set) == 1:
            output_missing_values = MissingValues(nan=False, none=True)
        else:
            # If there is more than one missing value, then it has to be
            # float('nan') or np.nan
            output_missing_values = MissingValues(nan=True, none=True)
    else:
        output_missing_values = MissingValues(nan=True, none=False)

    # create set without the missing values
    output = values - missing_values_set
    return output, output_missing_values


class _nandict(dict):
    """Dictionary with support for nans."""
    def __init__(self, mapping):
        super().__init__(mapping)
        for key, value in mapping.items():
            if is_scalar_nan(key):
                self.nan_value = value
                break

    def __missing__(self, key):
        if hasattr(self, 'nan_value') and is_scalar_nan(key):
            return self.nan_value
        raise KeyError(key)


def _map_to_integer(values, uniques):
    """Map values based on its position in uniques."""
    table = _nandict({val: i for i, val in enumerate(uniques)})
    return np.array([table[v] for v in values])


def _unique_python(values, *, return_inverse):
    # Only used in `_uniques`, see docstring there for details
    try:
        uniques_set = set(values)
        uniques_set, missing_values = _extract_missing(uniques_set)

        uniques = sorted(uniques_set)
        uniques.extend(missing_values.to_list())
        uniques = np.array(uniques, dtype=values.dtype)
    except TypeError:
        types = sorted(t.__qualname__
                       for t in set(type(v) for v in values))
        raise TypeError("Encoders require their input to be uniformly "
                        f"strings or numbers. Got {types}")

    if return_inverse:
        return uniques, _map_to_integer(values, uniques)

    return uniques


def _encode(values, *, uniques, check_unknown=True):
    """Helper function to encode values into [0, n_uniques - 1].
    Uses pure python method for object dtype, and numpy method for
    all other dtypes.
    The numpy method has the limitation that the `uniques` need to
    be sorted. Importantly, this is not checked but assumed to already be
    the case. The calling method needs to ensure this for all non-object
    values.
    Parameters
    ----------
    values : ndarray
        Values to encode.
    uniques : ndarray
        The unique values in `values`. If the dtype is not object, then
        `uniques` needs to be sorted.
    check_unknown : bool, default=True
        If True, check for values in `values` that are not in `unique`
        and raise an error. This is ignored for object dtype, and treated as
        True in this case. This parameter is useful for
        _BaseEncoder._transform() to avoid calling _check_unknown()
        twice.
    Returns
    -------
    encoded : ndarray
        Encoded values
    """
    if values.dtype.kind in 'OU':
        try:
            return _map_to_integer(values, uniques)
        except KeyError as e:
            raise ValueError(f"y contains previously unseen labels: {str(e)}")
    else:
        if check_unknown:
            diff = _check_unknown(values, uniques)
            if diff:
                raise ValueError(f"y contains previously unseen labels: "
                                 f"{str(diff)}")
        return np.searchsorted(uniques, values)


def _check_unknown(values, known_values, return_mask=False):
    """
    Helper function to check for unknowns in values to be encoded.
    Uses pure python method for object dtype, and numpy method for
    all other dtypes.
    Parameters
    ----------
    values : array
        Values to check for unknowns.
    known_values : array
        Known values. Must be unique.
    return_mask : bool, default=False
        If True, return a mask of the same shape as `values` indicating
        the valid values.
    Returns
    -------
    diff : list
        The unique values present in `values` and not in `know_values`.
    valid_mask : boolean array
        Additionally returned if ``return_mask=True``.
    """
    valid_mask = None

    if values.dtype.kind in 'UO':
        values_set = set(values)
        values_set, missing_in_values = _extract_missing(values_set)

        uniques_set = set(known_values)
        uniques_set, missing_in_uniques = _extract_missing(uniques_set)
        diff = values_set - uniques_set

        nan_in_diff = missing_in_values.nan and not missing_in_uniques.nan
        none_in_diff = missing_in_values.none and not missing_in_uniques.none

        def is_valid(value):
            return (value in uniques_set or
                    missing_in_uniques.none and value is None or
                    missing_in_uniques.nan and is_scalar_nan(value))

        if return_mask:
            if diff or nan_in_diff or none_in_diff:
                valid_mask = np.array([is_valid(value) for value in values])
            else:
                valid_mask = np.ones(len(values), dtype=bool)

        diff = list(diff)
        if none_in_diff:
            diff.append(None)
        if nan_in_diff:
            diff.append(np.nan)
    else:
        unique_values = np.unique(values)
        diff = np.setdiff1d(unique_values, known_values,
                            assume_unique=True)
        if return_mask:
            if diff.size:
                valid_mask = np.in1d(values, known_values)
            else:
                valid_mask = np.ones(len(values), dtype=bool)

        # check for nans in the known_values
        if np.isnan(known_values).any():
            diff_is_nan = np.isnan(diff)
            if diff_is_nan.any():
                # removes nan from valid_mask
                if diff.size and return_mask:
                    is_nan = np.isnan(values)
                    valid_mask[is_nan] = 1

                # remove nan from diff
                diff = diff[~diff_is_nan]
        diff = list(diff)

    if return_mask:
        return diff, valid_mask
    return diff

class _BaseEncoder(TransformerMixin, BaseEstimator):
    """
    Base class for encoders that includes the code to categorize and
    transform the input features.
    """

    def _check_X(self, X, force_all_finite=True):
        """
        Perform custom check_array:
        - convert list of strings to object dtype
        - check for missing values for object dtype data (check_array does
          not do that)
        - return list of features (arrays): this list of features is
          constructed feature by feature to preserve the data types
          of pandas DataFrame columns, as otherwise information is lost
          and cannot be used, eg for the `categories_` attribute.
        """
        if not (hasattr(X, 'iloc') and getattr(X, 'ndim', 0) == 2):
            # if not a dataframe, do normal check_array validation
            X_temp = check_array(X, dtype=None,
                                 force_all_finite=force_all_finite)
            if (not hasattr(X, 'dtype')
                    and np.issubdtype(X_temp.dtype, np.str_)):
                X = check_array(X, dtype=object,
                                force_all_finite=force_all_finite)
            else:
                X = X_temp
            needs_validation = False
        else:
            # pandas dataframe, do validation later column by column, in order
            # to keep the dtype information to be used in the encoder.
            needs_validation = force_all_finite

        n_samples, n_features = X.shape
        X_columns = []

        for i in range(n_features):
            Xi = self._get_feature(X, feature_idx=i)
            Xi = check_array(Xi, ensure_2d=False, dtype=None,
                             force_all_finite=needs_validation)
            X_columns.append(Xi)

        return X_columns, n_samples, n_features

    def _get_feature(self, X, feature_idx):
        if hasattr(X, 'iloc'):
            # pandas dataframes
            return X.iloc[:, feature_idx]
        # numpy arrays, sparse arrays
        return X[:, feature_idx]

    def _fit(self, X, handle_unknown='error', force_all_finite=True):
        X_list, n_samples, n_features = self._check_X(
            X, force_all_finite=force_all_finite)

        if self.categories != 'auto':
            if len(self.categories) != n_features:
                raise ValueError("Shape mismatch: if categories is an array,"
                                 " it has to be of shape (n_features,).")

        self.categories_ = []

        for i in range(n_features):
            Xi = X_list[i]
            if self.categories == 'auto':
                cats = _unique(Xi)
            else:
                cats = np.array(self.categories[i], dtype=Xi.dtype)
                if Xi.dtype.kind not in 'OU':
                    sorted_cats = np.sort(cats)
                    error_msg = ("Unsorted categories are not "
                                 "supported for numerical categories")
                    # if there are nans, nan should be the last element
                    stop_idx = -1 if np.isnan(sorted_cats[-1]) else None
                    if (np.any(sorted_cats[:stop_idx] != cats[:stop_idx]) or
                        (np.isnan(sorted_cats[-1]) and
                         not np.isnan(sorted_cats[-1]))):
                        raise ValueError(error_msg)

                if handle_unknown == 'error':
                    diff = _check_unknown(Xi, cats)
                    if diff:
                        msg = ("Found unknown categories {0} in column {1}"
                               " during fit".format(diff, i))
                        raise ValueError(msg)
            self.categories_.append(cats)

    def _transform(self, X, handle_unknown='error', force_all_finite=True):
        X_list, n_samples, n_features = self._check_X(
            X, force_all_finite=force_all_finite)

        X_int = np.zeros((n_samples, n_features), dtype=int)
        X_mask = np.ones((n_samples, n_features), dtype=bool)

        if n_features != len(self.categories_):
            raise ValueError(
                "The number of features in X is different to the number of "
                "features of the fitted data. The fitted data had {} features "
                "and the X has {} features."
                .format(len(self.categories_,), n_features)
            )

        for i in range(n_features):
            Xi = X_list[i]
            diff, valid_mask = _check_unknown(Xi, self.categories_[i],
                                              return_mask=True)

            if not np.all(valid_mask):
                if handle_unknown == 'error':
                    msg = ("Found unknown categories {0} in column {1}"
                           " during transform".format(diff, i))
                    raise ValueError(msg)
                else:
                    # Set the problematic rows to an acceptable value and
                    # continue `The rows are marked `X_mask` and will be
                    # removed later.
                    X_mask[:, i] = valid_mask
                    # cast Xi into the largest string type necessary
                    # to handle different lengths of numpy strings
                    if (self.categories_[i].dtype.kind in ('U', 'S')
                            and self.categories_[i].itemsize > Xi.itemsize):
                        Xi = Xi.astype(self.categories_[i].dtype)
                    else:
                        Xi = Xi.copy()

                    Xi[~valid_mask] = self.categories_[i][0]
            # We use check_unknown=False, since _check_unknown was
            # already called above.
            X_int[:, i] = _encode(Xi, uniques=self.categories_[i],
                                  check_unknown=False)

        return X_int, X_mask

    def _more_tags(self):
        return {'X_types': ['categorical']}
    
class OrdinalEncoder(_BaseEncoder):
    """
    Encode categorical features as an integer array.
    The input to this transformer should be an array-like of integers or
    strings, denoting the values taken on by categorical (discrete) features.
    The features are converted to ordinal integers. This results in
    a single column of integers (0 to n_categories - 1) per feature.
    Read more in the :ref:`User Guide <preprocessing_categorical_features>`.
    .. versionadded:: 0.20
    Parameters
    ----------
    categories : 'auto' or a list of array-like, default='auto'
        Categories (unique values) per feature:
        - 'auto' : Determine categories automatically from the training data.
        - list : ``categories[i]`` holds the categories expected in the ith
          column. The passed categories should not mix strings and numeric
          values, and should be sorted in case of numeric values.
        The used categories can be found in the ``categories_`` attribute.
    dtype : number type, default np.float64
        Desired dtype of output.
    handle_unknown : {'error', 'use_encoded_value'}, default='error'
        When set to 'error' an error will be raised in case an unknown
        categorical feature is present during transform. When set to
        'use_encoded_value', the encoded value of unknown categories will be
        set to the value given for the parameter `unknown_value`. In
        :meth:`inverse_transform`, an unknown category will be denoted as None.
        .. versionadded:: 0.24
    unknown_value : int or np.nan, default=None
        When the parameter handle_unknown is set to 'use_encoded_value', this
        parameter is required and will set the encoded value of unknown
        categories. It has to be distinct from the values used to encode any of
        the categories in `fit`. If set to np.nan, the `dtype` parameter must
        be a float dtype.
        .. versionadded:: 0.24
    Attributes
    ----------
    categories_ : list of arrays
        The categories of each feature determined during ``fit`` (in order of
        the features in X and corresponding with the output of ``transform``).
        This does not include categories that weren't seen during ``fit``.
    See Also
    --------
    OneHotEncoder : Performs a one-hot encoding of categorical features.
    LabelEncoder : Encodes target labels with values between 0 and
        ``n_classes-1``.
    Examples
    --------
    Given a dataset with two features, we let the encoder find the unique
    values per feature and transform the data to an ordinal encoding.
    >>> from sklearn.preprocessing import OrdinalEncoder
    >>> enc = OrdinalEncoder()
    >>> X = [['Male', 1], ['Female', 3], ['Female', 2]]
    >>> enc.fit(X)
    OrdinalEncoder()
    >>> enc.categories_
    [array(['Female', 'Male'], dtype=object), array([1, 2, 3], dtype=object)]
    >>> enc.transform([['Female', 3], ['Male', 1]])
    array([[0., 2.],
           [1., 0.]])
    >>> enc.inverse_transform([[1, 0], [0, 1]])
    array([['Male', 1],
           ['Female', 2]], dtype=object)
    """

    @_deprecate_positional_args
    def __init__(self, *, categories='auto', dtype=np.float64,
                 handle_unknown='error', unknown_value=None):
        self.categories = categories
        self.dtype = dtype
        self.handle_unknown = handle_unknown
        self.unknown_value = unknown_value

    def fit(self, X, y=None):
        """
        Fit the OrdinalEncoder to X.
        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data to determine the categories of each feature.
        y : None
            Ignored. This parameter exists only for compatibility with
            :class:`~sklearn.pipeline.Pipeline`.
        Returns
        -------
        self
        """
        if self.handle_unknown == 'use_encoded_value':
            if is_scalar_nan(self.unknown_value):
                if np.dtype(self.dtype).kind != 'f':
                    raise ValueError(
                        f"When unknown_value is np.nan, the dtype "
                        "parameter should be "
                        f"a float dtype. Got {self.dtype}."
                    )
            elif not isinstance(self.unknown_value, numbers.Integral):
                raise TypeError(f"unknown_value should be an integer or "
                                f"np.nan when "
                                f"handle_unknown is 'use_encoded_value', "
                                f"got {self.unknown_value}.")
        elif self.unknown_value is not None:
            raise TypeError(f"unknown_value should only be set when "
                            f"handle_unknown is 'use_encoded_value', "
                            f"got {self.unknown_value}.")

        self._fit(X)

        if self.handle_unknown == 'use_encoded_value':
            for feature_cats in self.categories_:
                if 0 <= self.unknown_value < len(feature_cats):
                    raise ValueError(f"The used value for unknown_value "
                                     f"{self.unknown_value} is one of the "
                                     f"values already used for encoding the "
                                     f"seen categories.")

        return self

    def transform(self, X):
        """
        Transform X to ordinal codes.
        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data to encode.
        Returns
        -------
        X_out : sparse matrix or a 2-d array
            Transformed input.
        """
        X_int, X_mask = self._transform(X, handle_unknown=self.handle_unknown)
        X_trans = X_int.astype(self.dtype, copy=False)

        # create separate category for unknown values
        if self.handle_unknown == 'use_encoded_value':
            X_trans[~X_mask] = self.unknown_value
        return X_trans

    def inverse_transform(self, X):
        """
        Convert the data back to the original representation.
        Parameters
        ----------
        X : array-like or sparse matrix, shape [n_samples, n_encoded_features]
            The transformed data.
        Returns
        -------
        X_tr : array-like, shape [n_samples, n_features]
            Inverse transformed array.
        """
        check_is_fitted(self)
        X = check_array(X, accept_sparse='csr')

        n_samples, _ = X.shape
        n_features = len(self.categories_)

        # validate shape of passed X
        msg = ("Shape of the passed X data is not correct. Expected {0} "
               "columns, got {1}.")
        if X.shape[1] != n_features:
            raise ValueError(msg.format(n_features, X.shape[1]))

        # create resulting array of appropriate dtype
        dt = np.find_common_type([cat.dtype for cat in self.categories_], [])
        X_tr = np.empty((n_samples, n_features), dtype=dt)

        found_unknown = {}

        for i in range(n_features):
            labels = X[:, i].astype('int64', copy=False)
            if self.handle_unknown == 'use_encoded_value':
                unknown_labels = labels == self.unknown_value
                X_tr[:, i] = self.categories_[i][np.where(
                    unknown_labels, 0, labels)]
                found_unknown[i] = unknown_labels
            else:
                X_tr[:, i] = self.categories_[i][labels]

        # insert None values for unknown values
        if found_unknown:
            X_tr = X_tr.astype(object, copy=False)

            for idx, mask in found_unknown.items():
                X_tr[mask, idx] = None

        return X_tr


from .columnselector import AccutuningColumnSelector  # noqa
from .columnlabeler import AccutuningColumnLabelEncoder  # noqa
from .integerencode import AccutuningIntegerEncode  # noqa
from .dtypeconvert import AccutuningDtypeConvert  # noqa
from .nullimputer import AccutuningNullImputerBycol  # noqa
from .outlier import AccutuningOutlierBycol  # noqa
from .powertransformer import AccutuningColTransformation  # noqa
from .sampler import AccutuningSampler  # noqa



from unittest import TestCase

from accutuning_helpers.text.tokenizer import KonlpyTokenizer, TwitterTokenizer


class TestKonlpyTokenizer(TestCase):

	def test_tokenize(self):
		# text = '아버지가방에들어가신다'
		text = '크리스토퍼 놀란에게 [우리는] 놀란다'
		tok = KonlpyTokenizer(tokenizer_name='Mecab')
		tokens = tok.tokenize(text)
		print(tokens)
		assert len(tokens) == 4 # ['크리스토퍼', '놀란', '우리', '놀란다']

		pos = tok.pos(text)
		print(pos)
		assert list(map(lambda x: x[1],pos)) == ['NNP', 'NNP', 'NP', 'VV+EC']


class TestTwitterTokenizer(TestCase):

	def test_tokenize(self):
		# text = '아버지가방에들어가신다'
		text = '크리스토퍼 놀란에게 [우리는] 놀란다'
		tok = TwitterTokenizer()
		# tokens = tok.stcs_to_words(text)
		tokens = tok.tokenize(text)
		print(tokens)
		assert len(tokens) == 6 # ['크리스토퍼', '놀란', '에게', '우리', '는', '놀란다']
		# assert len(tokens) == 4 # ['크리스토퍼', '놀란', '우리', '놀란다']

		# pos = tok.pos(text)
		# print(pos)
		# assert list(map(lambda x: x[1],pos)) == ['NNP', 'NNP', 'NP', 'VV+EC']


import logging
import math
from copy import copy
from datetime import datetime
from typing import Dict, List

import datasets
from flair.data import Corpus, Sentence, Tokenizer
from flair.datasets.document_classification import FlairDataset
from flair.models import TARSClassifier
from flair.optim import LinearSchedulerWithWarmup
from flair.tokenization import SegtokTokenizer
from flair.trainers import ModelTrainer
from torch.optim import AdamW

from accutuning_helpers.text.meta_learning import MetaLearner

logger = logging.getLogger(__name__)


def fetch(
		cls,
		**corpus_args
) -> Corpus:
	"""
	다루고 있는 한글 Dataset 종류
	1. KLUE benchmark (https://klue-benchmark.com/tasks) 중 - ynat, nli, sts
	-  Korean HateSpeech Dataset (https://github.com/kocohub/korean-hate-speech)
	"""

	# 1. fetch from huggingface
	dset_ = datasets.load_dataset(*cls.dataset_args)
	train_, dev_, test_ = dset_.get('train'), dset_.get('validation'), dset_.get('test')
	tokenizer = SegtokTokenizer()

	# 2. convert
	_train, _dev, _test = None, None, None
	if train_:
		_train = cls(dataset=train_, tokenizer=tokenizer)
	if dev_:
		_dev = cls(dataset=dev_, tokenizer=tokenizer)
	if test_:
		_test = cls(dataset=test_, tokenizer=tokenizer)

	# 3. bag in the corpus
	return Corpus(train=_train, dev=_dev, test=_test, name=corpus_args.get('name', _train.task_name), **corpus_args)


class HuggingfaceDataset(FlairDataset):
	def is_in_memory(self) -> bool:
		return True


class KlueYnatDataset(HuggingfaceDataset):
	dataset_args = ['klue', 'ynat']
	task_name = 'klue-ynat'
	label_name_map: Dict[int, str] = {
		0: "IT 과학",
		1: "경제",
		2: "사회",
		3: "생활 문화",
		4: "세계",
		5: "스포츠",
		6: "정치"
	}

	def __init__(
			self,
			dataset: datasets.Dataset,
			tokenizer: Tokenizer,
	):
		texts = dataset.data['title']  # features - [guid, title, label, url, date]
		tags = [self.label_name_map[i.as_py()] for i in dataset.data['label']]
		self._sentences: List[Sentence] = [
			Sentence(str(text), use_tokenizer=tokenizer).add_label(self.task_name, tag)
			for text, tag in zip(texts, tags)
		]

	def __getitem__(self, index: int) -> Sentence:
		return self._sentences[index]

	def __len__(self) -> int:
		return len(self._sentences)


class KlueNliDataset(HuggingfaceDataset):
	dataset_args = ['klue', 'nli']
	task_name = 'klue-nli'
	label_name_map: Dict[int, str] = {
		0: "유추 가능",
		1: "무관함",
		2: "모순",
	}

	def __init__(
			self,
			dataset: datasets.Dataset,
			tokenizer: Tokenizer,
	):
		self.sep = " [SEP] "

		# features - [guid, source, premise, hypothesis, label]
		prems, hypos = dataset.data['premise'], dataset.data['hypothesis']
		tags = [self.label_name_map[i.as_py()] for i in dataset.data['label']]
		self.pairs: List[Sentence] = [
			Sentence(text=str(prem) + self.sep + str(hypo), use_tokenizer=tokenizer).add_label(self.task_name, tag)
			for prem, hypo, tag in zip(prems, hypos, tags)
		]

	def __getitem__(self, index: int) -> Sentence:
		return self.pairs[index]

	def __len__(self) -> int:
		return len(self.pairs)


class KlueStsDataset(HuggingfaceDataset):
	dataset_args = ['klue', 'sts']
	task_name = 'klue-sts'
	label_name_map: Dict[int, str] = {
		0: "다른 의미",
		1: "같은 의미",
	}

	def __init__(
			self,
			dataset: datasets.Dataset,
			tokenizer: Tokenizer,
	):
		self.sep = " [SEP] "

		# features - [guid (string)	source (string)	sentence1 (string)	sentence2 (string)	labels (json)]
		# klue-sts-v1_train_00000
		# airbnb-rtt
		# 숙소 위치는 찾기 쉽고 일반적인 한국의 반지하 숙소입니다.
		# 숙박시설의 위치는 쉽게 찾을 수 있고 한국의 대표적인 반지하 숙박시설입니다.
		# { "label": 3.7, "real-label": 3.714285714285714, "binary-label": 1 }
		s1s, s2s = dataset.data['sentence1'], dataset.data['sentence2']
		tags = [self.label_name_map[d['binary-label'].as_py()] for d in dataset.data['labels']]
		self.pairs: List[Sentence] = [
			Sentence(text=str(s1) + self.sep + str(s2), use_tokenizer=tokenizer).add_label(self.task_name, tag)
			for s1, s2, tag in zip(s1s, s2s, tags)
		]

	def __getitem__(self, index: int) -> Sentence:
		return self.pairs[index]

	def __len__(self) -> int:
		return len(self.pairs)


class PawsXDataset(HuggingfaceDataset):
	dataset_args = ['paws-x', 'ko']
	task_name = 'paws-x'
	label_name_map: Dict[int, str] = {
		0: "전혀 다른 의미의 문장",
		1: "동일한 의미의 문장",
	}

	def __init__(
			self,
			dataset: datasets.Dataset,
			tokenizer: Tokenizer,
	):
		self.sep = " [SEP] "

		# features - [id, sentence1, sentence2, label]
		s1s, s2s = dataset.data['sentence1'], dataset.data['sentence2']
		tags = [self.label_name_map[i.as_py()] for i in dataset.data['label']]
		self.pairs: List[Sentence] = [
			Sentence(text=str(s1) + self.sep + str(s2), use_tokenizer=tokenizer).add_label(self.task_name, tag)
			for s1, s2, tag in zip(s1s, s2s, tags)
		]

	def __getitem__(self, index: int) -> Sentence:
		return self.pairs[index]

	def __len__(self) -> int:
		return len(self.pairs)


class NaverSentimentMovieCommentsDataset(HuggingfaceDataset):
	dataset_args = ['nsmc']
	task_name = 'nsmc'
	label_name_map: Dict[int, str] = {
		0: "부정적인 평가",
		1: "긍정적인 평가",
	}

	def __init__(
			self,
			dataset: datasets.Dataset,
			tokenizer: Tokenizer,
	):
		# features - [id, document, label]
		texts = dataset.data['document']
		tags = [self.label_name_map[i.as_py()] for i in dataset.data['label']]
		self._sentences: List[Sentence] = [
			Sentence(str(text).strip(), use_tokenizer=tokenizer).add_label(self.task_name, tag)
			for text, tag in zip(texts, tags) if str(text).strip()
		]

	def __getitem__(self, index: int) -> Sentence:
		return self._sentences[index]

	def __len__(self) -> int:
		return len(self._sentences)


class KoreanRestaurantReviewsDataset(HuggingfaceDataset):
	dataset_args = ["Wittgensteinian/KR3"]
	task_name = 'kr3'
	label_name_map: Dict[int, str] = {
		0: "부정적인 평가",
		1: "긍정적인 평가",
		2: "이도 저도 아닌 애매한 평가"
	}

	def __init__(
			self,
			dataset: datasets.Dataset,
			tokenizer: Tokenizer,
	):
		# features - [Rating, Review, __index_level_0__ ]
		texts = dataset.data['Review']
		tags = [self.label_name_map[i.as_py()] for i in dataset.data['Rating']]
		self._sentences: List[Sentence] = [
			Sentence(str(text), use_tokenizer=tokenizer).add_label(self.task_name, tag)
			for text, tag in zip(texts, tags)
		]

	def __getitem__(self, index: int) -> Sentence:
		return self._sentences[index]

	def __len__(self) -> int:
		return len(self._sentences)


class BaseMetaLearner(MetaLearner):

	def __init__(self, *args, base_language='ko', **kwargs):
		super(BaseMetaLearner, self).__init__(*args, **kwargs)
		self._lang = base_language

	def base_learning(
			self,
			embedding: str = 'klue/bert-base',
			down_sample: float = 1.0,
			sample_missing_splits=False,
			corpus_iteration: int = 3,
	):
		assert not self._tars_model and 0 < down_sample <= 1

		corpora = [
			fetch(KlueYnatDataset, sample_missing_splits=sample_missing_splits),
			fetch(KlueNliDataset, sample_missing_splits=sample_missing_splits),
			fetch(KlueStsDataset, sample_missing_splits=sample_missing_splits),
			fetch(PawsXDataset, sample_missing_splits=sample_missing_splits),
			fetch(NaverSentimentMovieCommentsDataset, sample_missing_splits=sample_missing_splits),
			fetch(KoreanRestaurantReviewsDataset, sample_missing_splits=sample_missing_splits),
			# NEWSGROUPS(sample_missing_splits=sample_missing_splits)
		]

		tars = TARSClassifier(
			embeddings=embedding,
		)
		# optimizer_params
		_params = list(tars.tars_model.named_parameters())
		no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
		decay = 0.01
		params = [
			{'params': [p for n, p in _params if not any(nd in n for nd in no_decay)], 'weight_decay': decay},
			{'params': [p for n, p in _params if any(nd in n for nd in no_decay)], 'weight_decay': 0.0}
		]
		optimizer = AdamW(params, lr=self._learning_rate, weight_decay=decay)
		# optimizer = Adam
		results = []
		mmdd = datetime.now().strftime("%m%d_%H%M")
		for i in range(1, corpus_iteration + 1):
			for c in corpora:
				if 0 < down_sample < 1.0:
					c = copy(c).downsample(percentage=down_sample)

				logger.info(f" start training for corpus {c.name}, {i} -- iteration")
				# tensorboard log directory
				log_dir = self._output_path / 'tensorboard' / mmdd / f'{c.name}_{i}'
				log_dir.mkdir(parents=True, exist_ok=True)

				if c.name in tars.list_existing_tasks():
					tars.switch_to_task(c.name)
				else:
					label_dict = c.make_label_dictionary(c.name)
					tars.add_and_switch_to_new_task(
						task_name=c.name,
						label_dictionary=label_dict,
						label_type=c.name,
						multi_label=label_dict.multi_label,
					)

				# initialize the text classifier trainer with corpus
				total_steps = math.ceil(len(c.train) / self._mini_batch_size) * self._max_epochs
				scheduler = LinearSchedulerWithWarmup(
					optimizer=optimizer,
					num_train_steps=total_steps,
					num_warmup_steps=self._warmup_fraction * total_steps,
				)

				trainer = ModelTrainer(tars, c)
				result = trainer.train(
					base_path=self._output_path / mmdd / c.name,  # path to store the model artifacts
					learning_rate=self._learning_rate,  # use very small learning rate
					# optimizer=AdamW,
					# optimizer=Adam, # default SGD
					optimizer=optimizer,
					scheduler=scheduler,
					mini_batch_size=self._mini_batch_size,  # small mini-batch size since corpus is tiny
					patience=self._patience,
					warmup_fraction=self._warmup_fraction,
					max_epochs=self._max_epochs,  # terminate after 10 epochs
					train_with_dev=self._train_with_dev,
					use_tensorboard=True,
					tensorboard_log_dir=log_dir,
				)
				results.append(result)

		self._tars_model = tars  # replace with fine tuned model
		logger.info(f'fine tuning completed for corpora:{[c.name for c in corpora]}, results:{results}')
		return results


if __name__ == "__main__":
	meta = BaseMetaLearner(
		model_path=None,  # base learning
		max_epochs=20,
		mini_batch_size=16,
		mini_batch_chunk_size=4,
		# learning_rate=1e-4,
		learning_rate=7e-5,
		# learning_rate=5e-5,  # learning rate
		# learning_rate=5e-3,
		# learning_rate=0.02,
		train_with_dev=False,
		base_language='ko'
	)
	# result = meta.base_learning(down_sample=1.0, embedding="kykim/bert-kor-base")
	result = meta.base_learning(down_sample=0.1, embedding="kykim/electra-kor-base")
	# result = meta.base_learning(down_sample=0.5, embedding="klue/bert-base")
	# result = meta.base_learning(down_sample=0.1, embedding="bert-base-cased")
	path = meta.save_model()
	print(path)



from abc import abstractmethod, ABC
from typing import List, Iterable, Union

import numpy as np
import pandas as pd
from flair.data import Tokenizer, Sentence
from sklearn.base import BaseEstimator, TransformerMixin


class TokenEmbedderBase(BaseEstimator, TransformerMixin, ABC):
	"""
	abstract class - 다양한 text embedder class를 추상화하는 class
	"""

	def __init__(self, feature_name: str):
		self.feature_name = feature_name

	@property
	@abstractmethod
	def tokenizer(self) -> Tokenizer:
		"""
			Embedder가 사용한 tokenizer를 반환. property 방식으로 접근
		Returns
		-------
			flair.data.Tokenizer
		"""
		pass

	def _add_embeddings_internal(self, sentences: List[Sentence]):
		pass

	def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
		vector_df = self.vectorize(X)
		X = X.drop(self.feature_name, axis=1)
		return pd.concat(
			[X, vector_df],
			axis=1
		)

	def vectorize(self, df: pd.DataFrame) -> pd.DataFrame:
		np.set_printoptions(threshold=100)
		corpus_embeddings = self.encode(df[self.feature_name].tolist())
		if hasattr(corpus_embeddings, 'A'):
			corpus_embeddings = corpus_embeddings.A  # 변환 csr -> ndarray
		columns = [
			f'{self.feature_name}_{i}'
			for i in range(corpus_embeddings.shape[1])
		]
		return pd.DataFrame(
			corpus_embeddings,
			columns=columns
		)

	@abstractmethod
	def encode(self, sentences: Iterable[str]) -> Union[np.ndarray, List[np.ndarray]]:
		pass



from sklearn.base import BaseEstimator, TransformerMixin
import pandas as pd

##########################################################################
# CTGAN을 기반으로 데이터를 Augmentation 하여 새로운 데이터셋을 제공합니다.
# use_class_balancer일 경우 target_column의 각 class 별 row 수를 맞춥니다.
##########################################################################
class AccutuningSampler(BaseEstimator, TransformerMixin):
    def __init__(self, sample=0, epochs=5, discrete_columns=None, target_column=None, use_class_balancer = False):
        self.sample = sample
        self.epochs = epochs
        self.discrete_columns = discrete_columns
        self.target_column = target_column
        self.use_class_balancer = use_class_balancer
        self.max_row = 0
        self.ctgans = []

    def fit(self, X, y=0, **fit_params):
        from ctgan import CTGANSynthesizer
        if self.use_class_balancer:
            for cls in X[self.target_column].unique():
                ctgan = CTGANSynthesizer()
                new_X = X[(X[self.target_column] == cls)]
                if not self.sample and self.max_row < len(new_X):
                    self.max_row = len(new_X)
                ctgan.fit(new_X, self.discrete_columns, epochs=self.epochs)
                self.ctgans.append(ctgan)

        else:
            ctgan = CTGANSynthesizer()
            ctgan.fit(X, self.discrete_columns, epochs=self.epochs)
            self.ctgans.append(ctgan)

        return self

    def transform(self, X, y=0):
        if self.use_class_balancer:
            result = pd.DataFrame()
            if self.sample and self.sample > 0:
                class_num = len(self.ctgans)
                remainder = self.sample % class_num
                for ctgan in self.ctgans:
                    if remainder > 0:
                        tmp = ctgan.sample(int(self.sample / class_num) + 1)
                        remainder = remainder - 1
                    else:
                        tmp = ctgan.sample(int(self.sample / class_num))
                    result = result.append(tmp, ignore_index=True)
            else:
                for ctgan in self.ctgans:
                    tmp = ctgan.sample(self.max_row)
                    result = result.append(tmp, ignore_index=True)
            return result

        else:
            return self.ctgans[0].sample(self.sample)



from sklearn.base import BaseEstimator, TransformerMixin
from scipy import sparse
from .onehotencoder_support import OneHotEncoder as OneHotEncoder_

import numpy as np

##########################################################################
# 전체 categorical_features에 대하여 One Hot Encoding을 실시합니다.
# Advanced Settings의 One Hot Encoding을 True로 설정할 경우 사용됩니다.
# 사용자가 지정한 Column에 대해서만 OHE를 실시하는 AccutuningCategoryConverter와는 구분됩니다.
##########################################################################
class AccutuningOneHotEncoder(BaseEstimator, TransformerMixin):
    def __init__(self, use_minimum_fraction=True, minimum_fraction=0.01,
                 categorical_features=None, random_state=None):

        self.use_minimum_fraction = use_minimum_fraction
        self.minimum_fraction = minimum_fraction
        self.categorical_features = categorical_features
        self.random_state = random_state

    def _fit(self, X, y=None):
        if self.use_minimum_fraction is False:
            self.minimum_fraction = None
        else:
            self.minimum_fraction = float(self.minimum_fraction)

        if self.categorical_features is not None:
            categorical_features = list(
                X.select_dtypes(include=['category', 'object']).columns
            )
        else:
            categorical_features = []

        self.preprocessor = OneHotEncoder_(
            minimum_fraction=self.minimum_fraction,
            categorical_features=categorical_features,
            sparse=True
        )

        return self.preprocessor.fit_transform(X)

    def fit(self, X, y=None):
        self._fit(X, y)
        return self

    def fit_transform(self, X, y=None):
        is_sparse = sparse.issparse(X)
        X = self._fit(X)
        if is_sparse:
            return X
        elif isinstance(X, np.ndarray):
            return X
        else:
            return X.toarray()

    def transform(self, X):
        is_sparse = sparse.issparse(X)
        if self.preprocessor is None:
            raise NotImplementedError()
        X = self.preprocessor.transform(X)
        if is_sparse:
            return X
        elif isinstance(X, np.ndarray):
            return X
        else:
            return X.toarray()


# # Imputation
# class Imputation(BaseEstimator, TransformerMixin):
#     def __init__(self, strategy='median', random_state=None):
#         self.strategy = strategy

#     def fit(self, X, y=None):
#         import sklearn.impute

#         self.preprocessor = sklearn.impute.SimpleImputer(
#             strategy=self.strategy, copy=False)
#         self.preprocessor = self.preprocessor.fit(X)
#         return self

#     def transform(self, X):
#         if self.preprocessor is None:
#             raise NotImplementedError()
#         return self.preprocessor.transform(X)


# # Variance Threshold
# class Variance_Threshold(BaseEstimator, TransformerMixin):
#     def __init__(self, random_state=None):
#         # VarianceThreshold does not support fit_transform (as of 0.19.1)!
#         pass

#     def fit(self, X, y=None):
#         self.preprocessor = VarianceThreshold(
#             threshold=0.0
#         )
#         self.preprocessor = self.preprocessor.fit(X)
#         return self

#     def transform(self, X):
#         if self.preprocessor is None:
#             raise NotImplementedError()
#         return self.preprocessor.transform(X)



import json
import os
from unittest import TestCase

from accutuning_helpers.text import utils as labeler_utils
from accutuning_helpers.text.meta_learning import MetaLearner, WORKPLACE_PATH, WORKPLACE_HOME


class TestMetaLearner(TestCase):

	def setUp(self) -> None:
		self.meta = MetaLearner(
			model_path=os.path.join(WORKPLACE_PATH, 'output/tars-electra-ko-v1.pt'),
			max_epochs=1,
			train_with_dev=True,
		)

	# def test_fine_tuning(self):
	# 	assert False
	#
	def test_label_predict(self):
		conf = {
			"random_seed": 42,
			"source_data_fp": "data/nnst_lt_10.csv",
			# "source_data_fp": "data/nnst_lt_1990.csv",
			"samples_fp": "",
			"related_stcs": "",
			"correct": True,
			"class_nm_list": ["생활", "기술"],
			"target_column_nm": "stcs",
			"samples_target_column_nm": "내용",
			"samples_tag_column_nm": "라벨",
			"labeler_worker_type": "zsl"
		}
		meta = self.meta
		result = meta.label_predict(**conf)
		result.pop('texts')  # too long to print
		result['predictions'] = [p.value for p in result['predictions']]
		print(json.dumps(result, indent=2, ensure_ascii=False), )
		# assert all(map(lambda x: x in conf['class_nm_list'], result['predictions']))

		input_path = os.path.join(WORKPLACE_PATH, conf['source_data_fp'])
		target_df = labeler_utils.load(input_path)
		gold, pred = target_df['tags'], result['predictions']
		labeler_utils.evaluate(gold, pred)

	def test_save_results(self):
		conf = {
			"random_seed": 42,
			"source_data_fp": "data/nnst_lt_10.csv",
			# "source_data_fp": "data/nnst_lt_1990.csv",
			"samples_fp": "",
			"related_stcs": "",
			"correct": True,
			"class_nm_list": ["생활", "기술"],
			"target_column_nm": "stcs",
			"samples_target_column_nm": "내용",
			"samples_tag_column_nm": "라벨",
			"labeler_worker_type": "zsl"
		}
		meta = self.meta
		result = meta.label_predict(**conf)
		output_path_info = meta.save_result(
			result_csv_filename='test.csv',
			**result
		)
		assert 'execution_time' in output_path_info
		assert os.path.exists(os.path.join(WORKPLACE_HOME, output_path_info['output_fp']))
		print(json.dumps(output_path_info, indent=2, ensure_ascii=False), )

	def test_few_shot_learning(self):
		conf = {
			"random_seed": 42,
			# "source_data_fp": "data/nnst_lt_10.csv",
			"source_data_fp": "data/nnst_lt_1990.csv",
			"samples_fp": "data/nnst_lt_10.csv",
			"related_stcs": "",
			"correct": False,
			"class_nm_list": None,
			"target_column_nm": "stcs",
			"samples_target_column_nm": "stcs",
			"samples_tag_column_nm": "tags",
			"labeler_worker_type": "fsl",
		}
		meta = self.meta
		result = meta.few_shot_learning(**conf)
		result.pop('texts')  # too long to print
		result['predictions'] = [p.value for p in result['predictions']]
		# print(json.dumps(result, indent=2, ensure_ascii=False),)

		input_path = os.path.join(WORKPLACE_PATH, conf['source_data_fp'])
		target_df = labeler_utils.load(input_path)
		gold, pred = target_df['tags'], result['predictions']
		labeler_utils.evaluate(gold, pred)

		## TODO: correct labels - 적용
		# conf['correct'] = True
		# result = meta.few_shot_learning(**conf)
		# result['predictions'] = [p.value for p in result['predictions']]
		# pred = result['predictions']
		# labeler_utils.evaluate(gold, pred)
		assert result

	def test_zero_shot_learning(self):
		conf = {
			"random_seed": 42,
			"source_data_fp": "data/nnst_lt_10.csv",
			# "source_data_fp": "data/nnst_lt_1990.csv",
			"samples_fp": "",
			"related_stcs": "",
			"correct": True,
			"class_nm_list": ["생활", "기술"],
			"target_column_nm": "stcs",
			"samples_target_column_nm": "내용",
			"samples_tag_column_nm": "라벨",
			"labeler_worker_type": "zsl"
		}
		meta = self.meta
		result = meta.zero_shot_learning(**conf)
		result.pop('texts')  # too long to print
		result['predictions'] = [p.value for p in result['predictions']]
		print(json.dumps(result, indent=2, ensure_ascii=False), )
		# assert all(map(lambda x: x in conf['class_nm_list'], result['predictions']))

		input_path = os.path.join(WORKPLACE_PATH, conf['source_data_fp'])
		target_df = labeler_utils.load(input_path)
		gold, pred = target_df['tags'], result['predictions']
		labeler_utils.evaluate(gold, pred)


class TestMetaBaseLearner(TestCase):

	def setUp(self) -> None:
		self.meta = MetaLearner(
			model_path=None,
			max_epochs=10,
			train_with_dev=True,
		)

	def test_base_learning(self):
		result = self.meta.base_learning(down_sample=0.01)
		assert result
		path = self.meta.save_model()
		assert os.path.exists(path)



from sklearn.base import BaseEstimator, TransformerMixin
from datetime import datetime
import time
import pandas as pd
import dateutil.parser

##########################################################################
# Datetime을 표현하는 Column (feature_name)을 지정된 방식에 따라 변환하여 새 Column으로 추가합니다.
# datetime_format은 사용자가 입력한 컬럼의 표현 양식을 나타내며, 값이 입력되지 않아도 더 느리지만 추정하여 변환할 수 있습니다.
# populate_features일 경우 feature_name을 연, 월, 일, 시간, 분, 초 등의 새 Columns로 변환하여 추가합니다.
# convert_timestamp일 경우 feature_name을 정수형태의 timestamp로 변환한 column을 추가합니다. (ex.2015-07-31 -> 1438268400)
##########################################################################
class AccutuningDatetime64Converter(BaseEstimator, TransformerMixin):
    def __init__(self, feature_name, datetime_format=None, populate_features=False, convert_timestamp=False):
        self.feature_name = feature_name
        self.datetime_format = datetime_format
        self.populate_features = populate_features
        self.convert_timestamp = convert_timestamp

    def fit(self, X, y=0, **fit_params):
        return self

    def transform(self, X, y=0):
        cn = self.feature_name
        target_column = X.loc[:, cn]
        # X.loc[:, col] = pd.to_datetime(converting_col, format='%Y-%m-%dT%H:%M:%SZ', errors='coerce')
        if self.datetime_format:
            X.loc[:, cn] = pd.to_datetime(
                target_column,
                format=self.datetime_format,
                errors='coerce'
            )
            target_column = X.loc[:, cn]
        else:
            def _parse(x):
                try:
                    # TODO: 매번 format을 유추하기 때문에 느리다;
                    ret = dateutil.parser.parse(x)
                # except dateutil.parser.ParserError:
                #     ret = datetime.fromtimestamp(0)
                except Exception:
                    ret = datetime.fromtimestamp(0)
                return ret
            X.loc[:, cn] = target_column.map(lambda x: _parse(x))
            target_column = X.loc[:, cn]

        X_tr = X.copy()
        if self.populate_features:
            idx = X.columns.get_loc(cn)
            X_tr.insert(idx + 1, cn + '_year', X[cn].dt.year)
            X_tr.insert(idx + 2, cn + '_month', X[cn].dt.month)
            X_tr.insert(idx + 3, cn + '_week', X[cn].dt.week)
            X_tr.insert(idx + 4, cn + '_day', X[cn].dt.day)
            X_tr.insert(idx + 5, cn + '_hour', X[cn].dt.hour)
            X_tr.insert(idx + 6, cn + '_minute', X[cn].dt.minute)
            X_tr.insert(idx + 7, cn + '_second', X[cn].dt.second)
            X_tr.insert(idx + 8, cn + '_dayofweek', X[cn].dt.dayofweek)
            X_tr.drop(columns=[cn], inplace=True)

        if self.convert_timestamp:
            X_tr.loc[:, cn] = target_column.map(lambda x: time.mktime(
                x.timetuple()
            ))

        if not self.populate_features and not self.convert_timestamp:
            X_tr.loc[:, cn] = target_column.astype('object')

        return X_tr



from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import LabelEncoder
from sklearn.utils.multiclass import type_of_target

##########################################################################
# Classification 문제에 해당할 경우 Target column을 int로 변환합니다.
##########################################################################
class AccutuningColumnLabelEncoder(BaseEstimator, TransformerMixin):
    def __init__(self, column_name):
        self.column_name = column_name
        self.le = None

    def fit(self, X, y=0, **fit_params):
        if self.column_name in X.columns:
            col = X.loc[:, self.column_name]
            try:
                task_type = type_of_target(col)
            except Exception:
                task_type = 'error'

            # if task_type == 'binary' or task_type.startswith('multiclass'):
            if task_type == 'binary' or col.dtype == 'object':
                self.le = LabelEncoder()
                self.le.fit(col)
        return self

    def transform(self, X, y=0):
        X_tr = X.copy()
        if self.le and self.column_name in X.columns:
            X_tr.loc[:, self.column_name] = self.le.transform(X.loc[:, self.column_name])
        return X_tr



import numpy as np


class EnsembleProxy(object):
    def __init__(self, objects_with_weight):
        self.models = [
            e[1]
            for e in objects_with_weight
        ]
        self.weights_ = [
            e[0]
            for e in objects_with_weight
        ]

    def predict(self, X):
        all_predictions = [
            model.predict(X.copy())  # predict_proba
            for model in self.models
        ]
        if len(all_predictions) > 1:
            return self.ensemble_predictions(all_predictions)
        else:
            return all_predictions[0]

    def predict_proba(self, X):
        # TODO: 모든 모델에 대해서 수행하도록; using np.average with weights
        prediction = self.models[0].predict_proba(X.copy())  # predict_proba
        return prediction

    def ensemble_predictions(self, predictions):
        predictions = np.asarray(predictions)

        if predictions.shape[0] == len(self.weights_):
            return np.average(predictions, axis=0, weights=self.weights_)

        # predictions do not include those of zero-weight models.
        elif predictions.shape[0] == np.count_nonzero(self.weights_):
            non_null_weights = [w for w in self.weights_ if w > 0]
            return np.average(predictions, axis=0, weights=non_null_weights)

        # If none of the above applies, then something must have gone wrong.
        else:
            raise ValueError("The dimensions of ensemble predictions"
                             " and ensemble weights do not match!")



from accutuning_helpers.text.meta_corpus import KlueStsDataset, fetch


def test_fetch_klue_sts():
	corpus = fetch(KlueStsDataset, sample_missing_splits=False)
	print(corpus)
	assert len(corpus.train) == 11668



from sklearn.base import BaseEstimator, TransformerMixin
from pandas.core.dtypes.common import is_numeric_dtype
import logging

##########################################################################
# outlier_cols에 해당되는 columns에 대하여 Outlier 제거를 실시합니다.
# BOX_PLOT_RULE, Z_SCORE 두 가지 방법을 선택할 수 있습니다.
# fitted는 기능이 전처리 과정에서만 작동하고, test set이 해당 클래스를 거치며 row를 지우는 일이 없도록 하는 역할을 수행합니다.
##########################################################################
class AccutuningOutlierBycol(BaseEstimator, TransformerMixin):
    def __init__(self, outlier_cols, outlier_strategy, outlier_threshold=None, fitted=False):
        self.outlier_cols = outlier_cols
        self.outlier_strategy = outlier_strategy
        self.outlier_threshold = outlier_threshold
        self.fitted = fitted

    def fit(self, X, y=0, **fit_params):
        return self

    def transform(self, X, y=0):
        X_tr = X.copy()
        if self.fitted:
            logging.critical(
                'AccutuningOutlierBycol: It is already transformed and will not be transformed again (for prediction).'
            )
        else:
            for col in self.outlier_cols:
                try:
                    converting_col = X.loc[:, col]
                except KeyError:
                    logging.critical(
                        'AccutuningOutlierBycol: No such column name in the dataset {}'.format(col)
                    )
                else:
                    if self.outlier_strategy == 'BOX_PLOT_RULE':
                        if is_numeric_dtype(converting_col):
                            q1 = converting_col.quantile(0.25)
                            q3 = converting_col.quantile(0.75)
                            IQR = q3 - q1
                            X_tr = X_tr[(X_tr[col] >= (q1 - 1.5 * IQR)) & (X_tr[col] <= (q3 + 1.5 * IQR))]
                            deleted = X_tr.shape[0] - sum((X_tr[col] >= (q1 - 1.5 * IQR)) & (X_tr[col] <= (q3 + 1.5 * IQR)))
                            logging.debug(f'AccutuningOutlierBycol: {deleted} rows deleted for column {col}')
                    elif self.outlier_strategy == 'Z_SCORE':
                        if is_numeric_dtype(converting_col):
                            if self.outlier_threshold is None:
                                self.outlier_threshold = 3
                            z = (X_tr[col] - converting_col.mean()) / converting_col.std()
                            X_tr = X_tr[abs(z) <= self.outlier_threshold]
                            deleted = sum(abs(z) > self.outlier_threshold)
                            logging.debug(f'AccutuningOutlierBycol: {deleted} rows deleted for column {col}')
                    else:
                        logging.critical(
                            'Not a proper method'
                        )
        self.fitted = True
        return X_tr



import os
from time import perf_counter

import pandas as pd

from accutuning_helpers.text.embedder_bert import BERTVectorizer


def timer(fn):
	"""
	timer decorator
	"""

	def inner(*args, **kwargs):
		print(f'# Start {fn.__name__}')
		start_time = perf_counter()
		fn(*args, **kwargs)
		end_time = perf_counter()
		execution_time = end_time - start_time
		print(f'## {fn.__name__} took {execution_time:.5f}s to execute')

	return inner


class TestBERTVectorizer:

	def __init__(self):
		home = os.environ['ACCUTUNING_WORKSPACE']
		# datafile = os.path.join(home, 'data/naver_movie_comments_data_small.txt')
		# df = pd.read_csv(datafile, sep='\t', header=None, names=['movie_ids', 'comments', 'rates'])
		datafile = os.path.join(home, 'data/nnst_lt_1990.csv')
		self.df = pd.read_csv(datafile)
		self.col_name = 'stcs'
		self.embedder = BERTVectorizer(feature_name=self.col_name, batch_size=16)

	def test_tokenize(self):
		assert len(self.df) > 0
		comments = self.df[self.col_name]
		comment = comments[0]

		tokens = self.embedder.tokenizer.tokenize(comment)
		print(f'tokens:{tokens}')

	# assert tokens == ['크리스토퍼', '놀란', '우리', '놀란', '다']

	@timer
	def test_fit(self):
		X = self.df
		self.embedder.fit(X)
		print(f'\nvocab lenghth:{self.embedder.embedding_length}')
		assert self.embedder.embedding_length == 512

	@timer
	def test_transform(self):
		X = self.df
		vec = self.embedder.transform(X)
		# print(f'vector:{vec}')
		print(f'vector.shape:{vec.shape}')
		# assert vec.shape[1] == 9994
		assert vec.shape[1] == 512 + 2

	@timer
	def test_fit_transform(self):
		X = self.df
		vec = self.embedder.fit_transform(X)
		print(f'\nvocab lenghth:{self.embedder.embedding_length}')
		# assert self.embedder.embedding_length == 9994
		assert self.embedder.embedding_length == 512
		print(f'vector:{vec}')
		print(f'vector.shape:{vec.shape}')


if __name__ == '__main__':
	test = TestBERTVectorizer()
	test.test_fit()
	test.test_transform()
	test.test_fit_transform()



import pandas as pd
import numpy as np
from rdkit import Chem
from rdkit.Chem import AllChem, QED, MolFromSmiles, MolFromInchi, MolToSmiles, Descriptors, MolSurf, rdMolDescriptors
from rdkit.Chem.AllChem import GetMorganFingerprintAsBitVect
from rdkit.Chem.Crippen import MolLogP 
from rdkit.Chem.QED import qed
from sklearn.base import BaseEstimator, TransformerMixin
import transformers
import logging

logger = logging.getLogger()

class AccutuningChemicalConverter(BaseEstimator, TransformerMixin):
    """
    Transform Chemical molecule string into vectors or tokens using Rdkit or transformers pretrained model.
    
    Parameters
    ----------
    feature_names : list
        List of feature names (str) to convert
    feature_format : str, default='smiles'
        Molecule formats, 'smiles' or 'inchi'
    how : str, default='ecfp'
        Method to convert the molecule string, 'ecfp', 'fcfp' or 'tokenize'
    dim : int, default=512
        Dimension for the molecule conversion
        Used only if 'how' is 'ecfp' or 'fcfp'
    add_properties : list, default=[]
        Add chemical properties of the input molecule columns for each.
        'QED', 'MW', 'ALOGP', 'HBA', 'HBD', 'PSA' ...
    """

    def __init__(self, feature_names, feature_format='smiles', how='ecfp', dim=512, add_properties=None):
        self.feature_names = feature_names
        self.feature_format = feature_format
        self.how = how
        self.dim = dim
        self.add_properties = add_properties

    def fit(self, X, y=0, **fit_params):
        
        if self.how == 'tokenize': 
            from transformers import AutoTokenizer
            self.converter = AutoTokenizer.from_pretrained('seyonec/ChemBERTa_zinc250k_v2_40k', use_fast=True)
            self.pad = self.converter.pad_token_id
        else:
            self.converter = self.how
        
        return self

    def transform(self, X, y=0):
        X_tr = X.copy()
        for feature in self.feature_names:
            target_column = X.loc[:, feature]
            target_idx = X.columns.get_loc(feature)
            X_front = X.columns[:target_idx]
            X_back = X.columns[target_idx + 1:]
        
            converted = target_column.map(lambda x: self._convert_mol(x, 
                                                                self.feature_format, 
                                                                self.converter,
                                                                self.dim))
            
            f = lambda x: '{}_{}'.format(feature, x)            
            converted = pd.DataFrame(converted.tolist()).rename(columns=f)
            if converted.isnull().values.any():
                converted.fillna(self.pad, inplace=True, downcast='infer')

            X_tr = pd.concat([X_tr[X_front], converted, X_tr[X_back]], axis=1)
            
            if self.add_properties:
                props = target_column.map(lambda x: self._add_properties(x, 
                                                                    self.feature_format,
                                                                    self.add_properties))
                X_tr = X_tr.join(pd.DataFrame.from_records(props, columns=self.add_properties))

        return X_tr

    def _convert_mol(self, sequence, feature_format, how, dim):
        if feature_format == 'smiles':
            m = MolFromSmiles(sequence)
        elif feature_format == 'inchi':
            m = MolFromInchi(sequence)
        else:
            logger.critical('AutoinsightChemicalConverter: Not a valid Molecule format selected.')
            m = None
        
        if how == 'ecfp':
            try:
                vec = GetMorganFingerprintAsBitVect(m, 2, nBits = dim)
                vec = vec.ToBitString()
            except Exception as e:
                vec = [0] * self.dim
        elif how == 'fcfp':
            try:
                vec = GetMorganFingerprintAsBitVect(m, 2, useFeatures = True, nBits = dim)
                vec = vec.ToBitString()
            except Exception as e:
                vec = [0] * self.dim
        elif isinstance(how, transformers.RobertaTokenizerFast):
            smiles = MolToSmiles(m)
            vec = how(smiles)['input_ids']
        else:
            logger.critical('AutoinsightChemicalConverter: No proper conversion method provided')
        
        converted = list(vec)
        
        return converted
    
    def _add_properties(self, sequence, feature_format, properties):
        props = []
        
        if feature_format == 'smiles':
            m = MolFromSmiles(sequence)
        elif feature_format == 'inchi':
            m = MolFromInchi(sequence)
        else:
            logger.critical('AutoinsightChemicalConverter: Not a valid Molecule format selected.')
            m = None
        
        try:
            pp = QED.properties(m)
            if properties:
                for i in properties:
                    if i == 'MW': 
                        props.append(round(pp.MW, 2))
                    elif i == 'ALOGP':
                        props.append(round(pp.ALOGP, 2))
                    elif i == 'HBA':
                        props.append(round(pp.HBA, 2))
                    elif i == 'HBD':
                        props.append(round(pp.HBD, 2))
                    elif i == 'PSA':
                        props.append(round(pp.PSA, 2))
                    elif i == 'QED':
                        props.append(round(QED.qed(m), 2))
                    else:
                        logger.critical(f'AutoinsightChemicalConverter: Not a property available: {i}')
        except Exception as e:
            props = [0] * len(properties)
            logger.critical(f'AutoinsightChemicalConverter: Fail to get properties from smiles: {mol}')
            logger.critical(e, exc_info=True)

        return props



from sklearn.base import BaseEstimator, TransformerMixin
from pandas.core.dtypes.common import is_numeric_dtype
import logging
import numpy as np
import pandas as pd
from scipy import stats

##########################################################################
# Column 순서대로 제시되는 strategies에 따라 feature transformation을 실시합니다.
# 각 Strategy에 따라 변환 가능한 데이터의 제한이 존재하고, 이럴 경우 다른 strategy로 변경됩니다.
##########################################################################
class AccutuningColTransformation(BaseEstimator, TransformerMixin):
    def __init__(self, strategies):
        # self.trans_cols = trans_cols
        self.strategies = strategies

    def fit(self, X, y=0, **fit_params):
        new_col, new_strategies = [], []
        for (col, strategy) in self.strategies:
            try:
                converting_col = X.loc[:, col]
            except KeyError:
                logging.critical(
                    f'Power transformer: No such column name in the dataset. {col}'
                )
            else:
                if is_numeric_dtype(converting_col):
                    if strategy == 'LOG':
                        if any(x == 0 for x in converting_col.values):
                            logging.warning(
                                f'The column {converting_col.name} contains 0 which will be converted to -Inf. LOG Transformation is skipped.'
                            )
                        elif any(x < 0 for x in converting_col.values):
                            logging.warning(
                                f'The column {converting_col.name} contains negative numbers which will be converted to nan. LOG Transformation is skipped.'
                            )
                        else:
                            new_strategies.append('LOG')
                            new_col.append(col)
                    elif strategy == 'SQUARED_ROOT':
                        if any(x < 0 for x in converting_col.values):
                            logging.warning(
                                f'The column {converting_col.name} contains negative numbers which will be converted to nan. SQUARED_ROOT Transformation is skipped'
                            )
                        else:
                            new_strategies.append('SQUARED_ROOT')
                            new_col.append(col)
                    elif strategy == 'SQUARE':
                        new_strategies.append('SQUARE')
                        new_col.append(col)
                    # box-cox transformation, yeo-johnson transformation
                    elif strategy == 'BOX_COX_TRANSFORMATION':
                        if any(x <= 0 for x in converting_col.values):
                            logging.warning(
                                f'Data must be positive to use Box-Cox Transformation. Yeo-Johnson Transformation will be used instead for the column {converting_col.name}.'
                            )
                            new_strategies.append('YEO_JOHNSON_TRANSFORMATION')
                            new_col.append(col)
                        else:
                            new_strategies.append('BOX_COX_TRANSFORMATION')
                            new_col.append(col)
                    elif strategy == 'YEO_JOHNSON_TRANSFORMATION':
                        new_strategies.append('YEO_JOHNSON_TRANSFORMATION')
                        new_col.append(col)
                    elif strategy == 'NONE':
                        pass
                    else:
                        logging.critical(
                            'Not a proper method'
                        )
        self.new_cols_strategies = list(zip(new_col, new_strategies))
        return self

    def transform(self, X, y=0):
        X_tr = X.copy()
        for (col, strategy) in self.new_cols_strategies:
            if col in X.columns:  # 존재하지않는 컬럼은 skip합니다
                logging.debug(col, strategy)
                converting_col = X.loc[:, col]
                if strategy == 'LOG':
                    newcol = np.log(converting_col)
                elif strategy == 'SQUARED_ROOT':
                    newcol = np.sqrt(converting_col)
                elif strategy == 'SQUARE':
                    newcol = np.square(converting_col)
                # box-cox transformation, yeo-johnson transformation
                elif strategy == 'BOX_COX_TRANSFORMATION':
                    try:
                        newcol, _ = stats.boxcox((converting_col + 1).astype(float))
                    except ValueError:
                        # https://stackoverflow.com/questions/62116192/valueerror-data-must-not-be-constant
                        # prediction에서 row가 1개일때는 valueerror를 피할 수 없음.
                        newcol = None
                elif strategy == 'YEO_JOHNSON_TRANSFORMATION':
                    newcol, _ = stats.yeojohnson(converting_col.astype(float))
                else:
                    newcol = None

                if newcol is not None:
                    X_tr[col] = newcol
        return X_tr



import logging
import pickle
from collections import Counter
from typing import List, Tuple

import cleanlab
import numpy as np
import pandas as pd
import torch
from flair.data import Label
from langid import langid
from sklearn import metrics
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.linear_model import LogisticRegression

from accutuning_helpers.text import NOT_CONFIDENT_TAG, MIN_LABEL_FREQUENCY
from accutuning_helpers.text.tokenizer import TwitterTokenizer

logger = logging.getLogger(__name__)


def tensor_to_array(vectors) -> np.ndarray:
	if torch.is_tensor(vectors):
		if vectors.requires_grad:
			vectors = vectors.detach()
		if vectors.is_cuda:
			vectors = vectors.cpu()
		vectors = vectors.numpy()
	return vectors


def load(filepath) -> pd.DataFrame:
	with open(filepath, 'rt') as f:
		if filepath.endswith('.csv'):
			df = pd.read_csv(filepath)
		elif filepath.endswith('.tsv'):
			df = pd.read_table(filepath)
		elif filepath.endswith('.xls') or filepath.endswith('.xlsx'):
			df = pd.read_excel(filepath)
		elif filepath.endswith('.txt'):
			texts = [line.strip() for line in f.readlines()]
			df = pd.DataFrame(texts, columns=['text'])
		elif filepath.endswith('.json'):
			df = pd.read_json(f, orient='table')
		elif filepath.endswith('.pkl'):
			df = pickle.load(f)
		else:
			raise Exception('unknown extension')

	return df


def identify_language(corpus_list: List[str], norm_probs=False) -> Tuple[List[str], List[float]]:
	identifier = langid.LanguageIdentifier.from_modelstring(langid.model, norm_probs=norm_probs)
	language = [identifier.classify(c)[0] for c in corpus_list]  # rank
	lang_id_result = dict(Counter(language))
	top_langs = list(lang_id_result)
	rst = list(lang_id_result.values())
	top_overall = [r / sum(rst) * 100 for r in rst]
	return top_langs, top_overall


def correct_label(texts: List[str], labels: List[Label]) -> Tuple[List[str], List[Label]]:
	top_langs, top_overall = identify_language(texts[:5])  # 최대 5문장만 보면 됨
	if top_langs[0] == 'ko' and top_overall[0] > 80:
		tokenized_stcs = [TwitterTokenizer().stcs_to_words(s) for s in texts]
	else:
		logger.debug(f"Identify Language - Top language:{top_langs[:1]}, Top overall:{top_overall[:1]}")
		tokenized_stcs = texts

	tfidf = TfidfVectorizer().fit(tokenized_stcs)
	embs = tfidf.transform(tokenized_stcs).toarray()

	tags = [x.value for x in labels]
	tags = sorted(list(set(tags)))
	label2int = {l: i for i, l in enumerate(tags)}
	int2label = {i: l for i, l in enumerate(tags)}
	int_labels = [label2int[t] for t in tags]
	int_y = int_labels

	X = np.array(embs)
	s = np.array(int_y)

	psx = cleanlab.latent_estimation.estimate_cv_predicted_probabilities(
		X, s, clf=LogisticRegression(max_iter=1000, multi_class='auto', solver='lbfgs'))
	pyx = psx
	logger.debug("Fetched probabilities for", pyx.shape[0], 'examples and', pyx.shape[1], 'classes.')

	# Estimate the confident joint, a proxy for the joint distribution of label noise.
	cj, cj_only_label_error_indices = cleanlab.latent_estimation.compute_confident_joint(
		s, pyx,
		return_indices_of_off_diagonals=True,
	)

	label_errors_idx = cleanlab.pruning.get_noise_indices(
		s=s,
		psx=pyx,
		confident_joint=cj,
		prune_method='both',
		sorted_index_method='normalized_margin',  # ['prob_given_label', 'normalized_margin']
	)
	replacements = []
	for idx in label_errors_idx:
		replacements.append(int2label[np.argmax(pyx[idx])])
	for (index, replacement) in zip(label_errors_idx, replacements):
		labels[index].value = replacement
	return texts, labels


def sampling(df: pd.DataFrame, tag_column_name: str, class_nm_list: List[str], n_samples: int) -> pd.DataFrame:
	replace = True  # with replacement
	fn = lambda obj: obj.loc[np.random.choice(obj.index, n_samples, replace), :]
	rst = df.loc[df[tag_column_name].isin(class_nm_list)]
	rst = rst.groupby(tag_column_name, as_index=False).apply(fn)
	return rst


def evaluate(gold: List[str], pred: List[str]) -> None:
	counter = Counter(pred)
	print(f'분류 못함: {counter[NOT_CONFIDENT_TAG]} / {len(gold)} 건 포함 metrics')
	print(metrics.classification_report(gold, pred))

	pred_expt = []
	tags_expt = []
	for idx, p in enumerate(pred):
		if p == NOT_CONFIDENT_TAG:
			pass
		else:
			pred_expt.append(pred[idx])
			tags_expt.append(gold[idx])
	print(f'분류 못함: {counter[NOT_CONFIDENT_TAG]} / {len(gold)} 건 제외 metrics')
	print(metrics.classification_report(tags_expt, pred_expt))


def is_confident_in_label_prediction(tags: List[str], class_name_list: List) -> bool:
	total_cnt = len(tags)  # 전체 레이블 문장 건수
	dict_cls_num = Counter(tags)  # 레이블 결과 건수

	prop = dict_cls_num[NOT_CONFIDENT_TAG] / total_cnt  # 전체 건수 대비 '분류못함' 레이블의 건수 비율:  0.7 이상이면 제로샷 결과만 리턴
	if prop > 0.7:  # 분류못함이 70% 이상
		return False

	if any(filter(lambda x: (x not in dict_cls_num) or (dict_cls_num[x] <= MIN_LABEL_FREQUENCY), class_name_list)):
		return False
	return True



from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import OneHotEncoder
import pandas as pd

##########################################################################
# feature_name column에 대하여 One Hot Encoding을 실시합니다.
# 선택된 column 마다 한번씩 해당 class를 사용하여 transform을 진행합니다.
##########################################################################
class AccutuningCategoryConverter(BaseEstimator, TransformerMixin):
    def __init__(self, feature_name):
        self.feature_name = feature_name
        self.ohe = OneHotEncoder(sparse=False, handle_unknown='ignore')

    def fit(self, X, y=0, **fit_params):
        self.ohe.fit(X[[self.feature_name]], y)
        return self

    def transform(self, X, y=0):
        self.target_idx = X.columns.get_loc(self.feature_name)
        self.X_front = X.columns[:self.target_idx]
        self.X_back = X.columns[self.target_idx + 1:]
        new_X = self.ohe.transform(X[[self.feature_name]])
        new_df = pd.DataFrame(new_X, columns=[
            self.feature_name + '_' + str(cn)
            for cn in self.ohe.categories_[0]
        ])
        new_df.index = X.index
        X_tr = pd.concat([X[self.X_front], new_df, X[self.X_back]], axis=1)
        return X_tr



import json
import logging
import os
import pickle
from pathlib import Path
from time import perf_counter
from typing import Dict, List, Union

import pandas as pd
from flair.data import Sentence, Corpus, Label
from flair.datasets import FlairDatapointDataset
from flair.models import TARSClassifier
from flair.trainers import ModelTrainer
from torch.optim.adamw import AdamW

from accutuning_helpers.text import NOT_CONFIDENT_TAG
from accutuning_helpers.text import utils as labeler_utils

logger = logging.getLogger(__name__)

META_MODEL_DIR = '/code/resources/labeler_models/'
META_MODEL_BIN_KO = 'tars-bert-ko-v1.pt'
META_MODEL_BIN_EN = 'tars-base-v8.pt'

DEFAULT_TAG_COLUMN_NAME = 'gold_tags'
WORKPLACE_HOME = os.environ.get('ACCUTUNING_WORKSPACE_ROOT', '/workspace')
WORKPLACE_PATH = os.environ['ACCUTUNING_WORKSPACE']
PREDICTION_SCORE_THRESHOLD = 0.7


def timer(fn):
	"""
	timer decorator
	"""

	def inner(*args, **kwargs):
		logger.info(f'# Start {fn.__name__}')
		start_time = perf_counter()
		to_execute = fn(*args, **kwargs)
		end_time = perf_counter()
		execution_time = end_time - start_time
		logger.info(f'## {fn.__name__} took {execution_time * 100:.3f}ms to execute')
		result: Dict = to_execute
		result['start_time'] = start_time
		result['end_time'] = end_time
		result['execution_time'] = execution_time
		return result

	return inner


def save_output_file(filepath: Path, obj) -> Union[str, Path]:
	filepath.write_bytes(
		pickle.dumps(obj)
	)
	return filepath


def _relative_to_workspace(filepath: Union[str, Path]) -> str:
	if filepath is None:
		return filepath

	if isinstance(filepath, str):
		filepath = Path(filepath)
	if str(filepath).startswith(META_MODEL_DIR):
		return str(filepath.relative_to(META_MODEL_DIR))
	return str(filepath.relative_to(WORKPLACE_HOME))


def to_predictions(sentences: List[Sentence]) -> List[Label]:
	predictions = []
	for sentence in sentences:
		number_list = [label.score for label in sentence.labels]
		max_value = max(number_list) if number_list else 0
		if len(number_list) > 0:
			# if max_value >= PREDICTION_SCORE_THREASHOLD:
			max_index = number_list.index(max_value)
			max_label = sentence.labels[max_index]
			predictions.append(max_label)
		else:
			not_confident = Label(sentence, value=NOT_CONFIDENT_TAG, score=0.5)
			predictions.append(not_confident)  ## TODO: NOT Confident 분기 정교화?
	return predictions


def get_task_name(labels: List[str]) -> str:
	"""
	##FIXME: tricky - 명시적으로 task name을 받는게 좋음
	특정 task name이 없는 경우, unique label 의 집합을 task name 으로 한다
	"""
	unique = sorted(set(labels))
	return '__'.join(unique)


class MetaLearner:

	def __init__(
			self,
			output_path: Path = None,
			model_path: Union[str, Path] = None,
			tars_model: TARSClassifier = None,
			learning_rate=5e-5,
			mini_batch_size=16,
			mini_batch_chunk_size=8,
			warmup_fraction=0.1,
			patience=10,
			max_epochs=10,
			train_with_dev=True,
			n_samples=5,
			prediction_batch_size=16,
	):
		self._learning_rate = learning_rate
		self._mini_batch_size = mini_batch_size
		self._mini_batch_chunk_size = mini_batch_chunk_size
		self._warmup_fraction = warmup_fraction
		self._patience = patience
		self._max_epochs = max_epochs
		self._train_with_dev = train_with_dev
		self._n_samples = n_samples
		self._prediction_batch_size = prediction_batch_size
		self._output_path = output_path or Path(WORKPLACE_PATH, 'output')
		self._model_path = model_path
		self._lang = None  # lazy identify
		self._tars_model: TARSClassifier = tars_model  # lazy loading

	@property
	def model_path(self):
		"""
		현재 self._tars_model 이 load 되었던 directory를 return 한다.
		load 이후 학습이 진행된 경우 weight가 달라졌으므로 이에 유의한다.
		"""
		return self._model_path

	def _load_model(self, model_path: str = None, lang: str = 'ko'):
		"""
		언어가 결정되는 시점은 data를 읽은 이후라, 늦게 모델을 load 한다
		"""
		model = self._tars_model
		if not model:
			if not model_path:
				if lang == 'ko':
					model_path = os.path.join(META_MODEL_DIR, META_MODEL_BIN_KO)
				else:  # english 취급
					model_path = os.path.join(META_MODEL_DIR, META_MODEL_BIN_EN)

			model = TARSClassifier.load(model_path)
			self._tars_model = model
			self._model_path = model_path
			self._lang = lang
		return model

	def _identify_language(self, texts: List[str]):
		lang = self._lang
		if not lang:
			langs, _ = labeler_utils.identify_language(texts)
			lang = langs[0] if langs else 'en'  # default
			self._lang = lang
		return lang

	@timer
	def fine_tuning(
			self,
			df: pd.DataFrame = None,
			text_column_name: str = None,
			tag_column_name: str = None,
			texts: List[str] = None,
			tags: List[str] = None,
			task_name: str = None,
	) -> Dict[str, str]:
		assert df is not None or texts is not None, "df or text should be provided."

		texts = texts or df[text_column_name].values.tolist()
		tags = tags or df[tag_column_name].values.tolist()
		task_name = task_name or get_task_name(tags)

		tr = [Sentence(text).add_label(task_name, tag) for text, tag in zip(texts, tags)]
		dataset = FlairDatapointDataset(tr)
		corpus = Corpus(train=dataset, sample_missing_splits=False)

		lang = self._identify_language(texts)
		tars = self._load_model(model_path=self.model_path, lang=lang)

		if task_name in tars.list_existing_tasks():
			tars.switch_to_task(task_name)
		else:
			label_dict = corpus.make_label_dictionary(task_name)
			tars.add_and_switch_to_new_task(
				task_name=task_name,
				label_dictionary=label_dict,
				label_type=task_name,
				multi_label=label_dict.multi_label,
			)

		# initialize the text classifier trainer with corpus
		trainer = ModelTrainer(tars, corpus)

		# train model
		result = trainer.fine_tune(
			base_path=self._output_path,  # path to store the model artifacts
			learning_rate=self._learning_rate,  # use very small learning rate
			optimizer=AdamW,
			param_selection_mode=True,
			mini_batch_size=self._mini_batch_size,  # small mini-batch size since corpus is tiny
			mini_batch_chunk_size=self._mini_batch_chunk_size,
			max_epochs=self._max_epochs,  # terminate after 10 epochs
			train_with_dev=self._train_with_dev
		)
		self._tars_model = tars  # replace with fine tuned model
		logger.info(f'fine tuning result:{result}')
		return result

	def _shot_learning(
			self,
			texts: List[str],
			class_nm_list: List[str] = None,
	) -> List[Label]:
		lang = self._identify_language(texts)
		tars = self._load_model(model_path=self.model_path, lang=lang)

		sentences = [Sentence(text) for text in texts]
		if class_nm_list:  # zero shot
			tars.predict_zero_shot(sentences, class_nm_list)
		else:
			# already tasks are added at the fine tuning time
			tars.predict(sentences, mini_batch_size=self._prediction_batch_size)

		return to_predictions(sentences)

	@timer
	def zero_shot_learning(
			self,
			class_nm_list: List[str],
			target_column_nm: str = None,
			source_data_fp: str = None,
			texts: List[str] = None,
			tag_column_nm: str = None,
			**config_kwargs,
	) -> Dict[str, Union[str, List[str], List[Label]]]:
		assert texts or source_data_fp, 'texts or source data file should be provided.'

		if not texts:
			input_path = os.path.join(WORKPLACE_HOME, source_data_fp)
			target_df = labeler_utils.load(input_path)
			texts = target_df[target_column_nm].values.tolist()
		predictions = self._shot_learning(texts, class_nm_list)

		return {
			'text_name': target_column_nm,
			'texts': texts,
			'tag_name': tag_column_nm,
			'predictions': predictions,  # value만, score 제외
			'class_nm_list': class_nm_list,
		}

	@timer
	def few_shot_learning(
			self,
			target_column_nm: str = None,
			tag_column_nm: str = None,
			source_data_fp: str = None,
			texts: List[str] = None,
			samples_target_column_nm: str = None,
			samples_tag_column_nm: str = None,
			samples_fp: str = None,
			sample_texts: List[str] = None,
			sample_tags: List[str] = None,
			task_name: str = None,
			correct: bool = False,
			**config_kwargs,
	) -> Dict[str, Union[str, List[str], List[Label]]]:
		assert sample_texts or samples_fp, "sample texts or sample file should be provided."

		# 1. fine tuning with samples
		if sample_texts and sample_tags:
			self.fine_tuning(texts=sample_texts, tags=sample_tags, task_name=task_name)
		else:
			sample_path = os.path.join(WORKPLACE_HOME, samples_fp)
			samples_df = labeler_utils.load(sample_path)
			self.fine_tuning(samples_df, samples_target_column_nm, samples_tag_column_nm, task_name=task_name)

		# 2. predict whole data
		assert texts or source_data_fp, 'texts or source data file should be provided.'
		if not texts:
			input_path = os.path.join(WORKPLACE_HOME, source_data_fp)
			target_df = labeler_utils.load(input_path)
			texts = target_df[target_column_nm].values.tolist()
		predictions = self._shot_learning(texts, class_nm_list=None)

		# 3. correct predictions if told to do so
		# FIXME: cleanlab 2.0 적용. 분류못함 삭제되도록 유도 적용
		# if correct:
		# 	logger.debug('FSL - Correcting labels')
		# 	texts, predictions = labeler_utils.correct_label(texts, predictions)

		# 4. return result
		return {
			'text_name': target_column_nm,
			'texts': texts,
			'tag_name': tag_column_nm,
			'predictions': predictions,  # value만, score 제외
		}

	@timer
	def label_predict(
			self,
			class_nm_list,
			target_column_nm,
			source_data_fp,
			**config_kwargs,
	) -> Dict[str, Union[str, List[str], List[Label]]]:
		result: Dict = self.zero_shot_learning(
			class_nm_list,
			target_column_nm,
			source_data_fp,
			**config_kwargs,
		)
		return result

	def save_model(self, file_path=None) -> str:
		file_path = file_path or str(self._output_path / f'fine_tuned_{self._lang}.pt')
		self._tars_model.save(model_file=file_path)
		logger.info(f'model saved in the path:{file_path}')
		return file_path

	def save_result(
			self,
			result_csv_filename: str,
			text_name: str,
			texts: List[str],
			tag_name: str,
			predictions: List[Label],
			execution_time: float,
			tags: List[Union[int, str]] = None,
			model_path: str = None,
			**config_kwargs,  # misc arguments - ignored
	) -> Dict[str, str]:
		output_path = self._output_path
		tag_name = tag_name or DEFAULT_TAG_COLUMN_NAME

		# save results
		result = {}
		result[text_name] = texts
		if tags:  # gold labels
			result[tag_name] = tags

		p_labels = [label.value for label in predictions]
		p_scores = [label.score for label in predictions]
		result[f'{tag_name}_predicted'] = p_labels
		result['confidence'] = p_scores

		result_df = pd.DataFrame(result)
		result_df.to_csv(output_path / result_csv_filename, index=False)

		labels_path = save_output_file(output_path / 'labels.pkl', (p_labels, p_scores))
		clusters_path = save_output_file(output_path / 'clusters.pkl', list(set(p_labels)))

		output_path_info = {
			'output_fp': _relative_to_workspace(output_path / result_csv_filename),
			'labels': _relative_to_workspace(labels_path),
			'clusters': _relative_to_workspace(clusters_path),
			'fine_tuned_model': _relative_to_workspace(model_path),
			'execution_time': execution_time,
		}

		# save output location
		(output_path / 'output.json').write_text(
			json.dumps(output_path_info)
		)
		return output_path_info



from sklearn.base import BaseEstimator, TransformerMixin
from .ordinalencoder_tmp import OrdinalEncoder

##########################################################################
# DataFrame 내 Category나 String column일 경우 해당 column들은 Integer로 변환합니다.
# train set에서 없었던 값일 경우 transform 시에 unknown_value (기본값=-1)로 대체됩니다.
##########################################################################
class AccutuningIntegerEncode(BaseEstimator, TransformerMixin):
    def __init__(self, unknown_value=-1):
        self.columns_to_encode = list()
        self.unknown_value = unknown_value
        self.oe = OrdinalEncoder(handle_unknown='use_encoded_value', unknown_value=self.unknown_value)

    def fit(self, X, y=0, **fit_params):
        self.columns_to_encode = list(
            X.select_dtypes(include=['object'])
        )

        for col in self.columns_to_encode:
            # Null 값이 있으면 encoder에서 에러 발생한다. (추후 수정 필요)
            # integer(float) or 'NaN'이 섞여있어도 에러 발생한다.
            X.loc[:, col] = X.loc[:, col].fillna('NaN').apply(str)

        try:
            self.oe.fit(X.loc[:, self.columns_to_encode])
        except Exception as e:
            print(self.columns_to_encode)
            print(e)
            raise
        return self

    def transform(self, X, y=0):
        X_tr = X.copy()
        for col in self.columns_to_encode:
            X_tr.loc[:, col] = X.loc[:, col].fillna('NaN').apply(str)
        X_tr.loc[:, self.columns_to_encode] = self.oe.transform(X_tr.loc[:, self.columns_to_encode]).astype('object')
        return X_tr



from .onehotencoder import AccutuningOneHotEncoder  # noqa
from .datetime64_converter import AccutuningDatetime64Converter  # noqa
from .category_converter import AccutuningCategoryConverter  # noqa
from .timeseries import AccutuningTimeseriesResample, AccutuningTimeseriesInterpolate # noqa
from .lag_column_adder import AccutuningLagColumnAdder  # noqa
# from .chemical import AccutuningChemicalConverter   # noqa



import os
from unittest import TestCase

import pandas as pd

from accutuning_helpers.preprocessing.dtypeconvert import AccutuningDtypeConvert

HOME = os.environ['ACCUTUNING_WORKSPACE']


class TestAccutuningDtypeConvert(TestCase):

	def test_transform_nnst(self):
		datafile = os.path.join(HOME, 'data/nnst_lt_10.csv')
		df = pd.read_csv(datafile)
		converter = AccutuningDtypeConvert(
			datatype_pair_match=[('Unnamed: 0', 'int64'), ('stcs', 'text'), ('tags', 'object')]
		)
		ret_df = converter.transform(X=df)
		assert len(ret_df.columns) == 76 # Unnamed, tags, stcs_vocabulary 74

	def test_transform_naver(self):
		datafile = os.path.join(HOME, 'data/naver_movie_comments_data_small.txt')
		df = pd.read_csv(datafile, sep='\t', names=['movie_ids', 'comments', 'rates'])
		converter = AccutuningDtypeConvert(
			datatype_pair_match=[('movie_ids', 'int64'), ('comments', 'text'), ('rates', 'object')]
		)
		ret_df = converter.transform(X=df)
		assert len(ret_df.columns) == 9996 # movie_id, rates, tf_vocabulary 9994



import logging
import math
from typing import Tuple, Literal, Dict, List
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin

logger = logging.getLogger(__name__)

TEXT_CONVERTER = Literal['tfidf', 'BERT', 'zero_shot', 'few_shot']
##########################################################################
# preprocessor 화면에서 받은 datatype_pair_match에 따라서 각 column들의 datatype을 변경합니다.
##########################################################################
class AccutuningDtypeConvert(BaseEstimator, TransformerMixin):
	def __init__(
			self,
			datatype_pair_match: List[Tuple[str, str]],
			text_converter: TEXT_CONVERTER = 'tfidf'
	):
		self.datatype_pair_match: Tuple[str, str] = datatype_pair_match
		self._converter = text_converter
		self._vector_dict: Dict[str, "TokenEmbedderBase"] = dict()
	def fit(self, X: pd.DataFrame, y=0, **fit_params) -> "AccutuningDtypeConvert":
		return self
	def transform(self, X: pd.DataFrame, y=0) -> pd.DataFrame:
		X_tr = X.copy()
		for (col, typ) in self.datatype_pair_match:
			try:
				converting_col = X_tr[col]
			except KeyError:
				logger.critical(
					'No such column name in the dataset - dtype convert'
				)
			else:
				if typ == 'text':
					ratio = 0.001
					min_df = min(math.floor(len(converting_col) * ratio), 10)
					vec = self._vectorizer_factory(col, min_df=min_df)
					if vec.embedding_length == 0:  # not fitted
						X_tr = vec.fit_transform(X_tr, y=y)
					else:
						X_tr = vec.transform(X_tr)
				elif str(converting_col.dtype) != typ:
					if typ == 'datetime64':
						pass
					elif typ == 'float64':
						# to_numeric으로 float 변환 안되는 것들을 NaN으로 치환 후 변환
						X_tr[col] = pd.to_numeric(X[col], errors='coerce').astype(typ)
					else:
						try:
							X_tr[col] = X[col].astype(typ)
						except ValueError:
							logger.critical(
								f'Failed to convert the datatype of column {col} to {str(typ)}. Set to default.'
							)
		return X_tr
	def _vectorizer_factory(self, col: str, **params) -> "TokenEmbedderBase":
		if col in self._vector_dict:
			return self._vector_dict[col]
		else:  # register vectorizer
			if self._converter == 'tfidf':
				from accutuning_helpers.text.embedder_tfidf import TfIdfTokenVectorizer
				vec = TfIdfTokenVectorizer(feature_name=col, **params)
			elif self._converter == 'BERT':
				## FIXME - flair version 에 맞게 huggingface version 4로 version up
				## FIXME - sentence transformer 기반 BERT vectorizer는 deprecated 될 예정
				from accutuning_helpers.text.embedder_bert import BERTVectorizer
				vec = BERTVectorizer(feature_name=col)
			self._vector_dict[col] = vec
			return vec



import os
import pickle
from unittest import TestCase

import pandas as pd

from accutuning_helpers.text.embedder_tfidf import TfIdfTokenVectorizer
from accutuning_helpers.text.tokenizer import KonlpyTokenizer


class TestTfidfTokenVectorizer(TestCase):

	def setUp(self) -> None:
		home = os.environ['ACCUTUNING_WORKSPACE']
		# datafile = os.path.join(home, 'data/naver_movie_comments_data_small.txt')
		# df = pd.read_csv(datafile, sep='\t', header=None, names=['movie_ids', 'comments', 'rates'])
		datafile = os.path.join(home, 'data/nnst_lt_1990.csv')
		# datafile = os.path.join(home, 'data/네이버영화평_sample.csv')
		df = pd.read_csv(datafile)
		self.train_df = df.iloc[:400]
		self.valid_df = df.iloc[400:]
		self.df = pd.concat([self.train_df, self.valid_df])

		self.tokenizer = KonlpyTokenizer(tokenizer_name='mecab')
		# self.embedder = TfIdfTokenVectorizer(feature_name='comments', tokenizer=self.tokenizer)
		self.embedder = TfIdfTokenVectorizer(
			feature_name='stcs',
			tokenizer=self.tokenizer,
			min_df=1,
			ngram_range=(1, 3),
			max_features=500
		)

	# self.embedder = TfIdfTokenVectorizer(feature_name='document', tokenizer=self.tokenizer, min_df=0)

	def test_tokenize(self):
		assert len(self.df) > 0
		# comments = self.df['comments']
		comments = self.df['stcs']
		# comments = self.df['document']
		comment = comments[0]

		tokens = self.tokenizer(comment)
		# assert tokens == ['크리스토퍼', '놀란', '우리', '놀란', '다']
		print(f'tokens:{tokens}')

	def test_fit(self):
		self.embedder.fit(self.df)
		print(f'\nvocab lenghth:{self.embedder.embedding_length}')
		# assert self.embedder.embedding_length == 9994
		# assert self.embedder.embedding_length == 74
		vocab = self.embedder._vectorizer.vocabulary_
		keywords = sorted(vocab.keys(), key=lambda x: vocab[x], reverse=True)
		print(keywords[:5])
		assert len(vocab.keys()) > 0

	def test_transform(self):
		if self.embedder.embedding_length == 0:
			self.embedder.fit(self.df)

		vec = self.embedder.transform(self.valid_df)
		print(f'vector:{vec}')
		print(f'vector.shape:{vec.shape}')
		# assert vec.shape[1] == 9994
		assert len(self.valid_df) == vec.shape[0]

	def test_fit_transform(self):
		X = self.df
		vec = self.embedder.fit_transform(X)
		print(f'\nvocab lenghth:{self.embedder.embedding_length}')
		# assert self.embedder.embedding_length == 9994
		# assert self.embedder.embedding_length == 74
		print(f'vector:{vec}')
		print(f'vector.shape:{vec.shape}')
		assert len(X) == vec.shape[0]

	def test_tfidf_max_features(self):
		if self.embedder.embedding_length == 0:
			self.embedder.fit(self.df)

		vec = self.embedder.transform(self.valid_df)
		assert self.embedder.embedding_length <= self.embedder._vectorizer.max_features

	def test_save(self):
		with open('test_dump.pkl', 'wb') as f:
			pickle.dump(self.embedder, f)
			print(f'located: {f}')

	def test_load(self):
		with open('test_dump.pkl', 'rb') as f:
			embedder = pickle.load(f)
			print(embedder)



from setuptools import setup, find_namespace_packages
setup(
    name="accutuning_helpers",
    version="1.0.32",
    packages=find_namespace_packages(),
    include_package_data=True,
    install_requires=[
        "numpy==1.19.5",
        "pandas==1.1.4",
        "scipy==1.5.2",
        "joblib==1.1.0",
        "scikit-learn==0.23.2",
        "xgboost==1.6.1",
        "lightgbm==3.3.2",
        "ctgan==0.4.3",
        "python-dateutil==2.8.2",
        # text added
        "konlpy==0.6",
        "customized_konlpy",
        "tweepy>=3.7.0",
        "langid>=1.1.6",
        "cleanlab>=1.0",
        "flair==0.11.3",
        "sentencepiece>=0.1.95",
        "datasets>=2.3.2", # Huggingface datasets
        "sentence-transformers>=2.2",
        # "protobuf==3.20.0",
        "conllu==4.4.1" # conllu== 4.4.2 로 update되면서 flair 11.3에서 문제 발생하여 hotfix 조치함. flair version up시 삭제할 것
    ]
)



import logging
from typing import Iterable, List

import numpy as np
import pandas as pd
from flair.data import Tokenizer
from sentence_transformers import SentenceTransformer

from accutuning_helpers.text.embedder import TokenEmbedderBase

logger = logging.getLogger(__name__)


class BERTVectorizer(TokenEmbedderBase):

	def __init__(
			self,
			feature_name,
			bert_model_name='sentence-transformers/distiluse-base-multilingual-cased-v1',
			batch_size=16,
	):
		super(BERTVectorizer, self).__init__(feature_name)

		# 학습하지 않은 BERT 대신 sentence transformers가 자체적으로 multilingual로 학습한 vector로 수정
		# word_embedding_model = models.Transformer(bert_model_name)
		# pooling_model = models.Pooling(
		# 	word_embedding_model.get_word_embedding_dimension(),
		# 	pooling_mode_mean_tokens=True,
		# 	pooling_mode_cls_token=False,
		# 	pooling_mode_max_tokens=False)
		# self.model = SentenceTransformer(modules=[word_embedding_model, pooling_model])
		self.model = SentenceTransformer(bert_model_name)
		self._batch_size = batch_size
		self._tokenizer = _Tokenizer(delegator=self.model)
		self._pool = self.model.start_multi_process_pool()
		logger.debug(f"SentenceTransformer start multi_process_pool")

	def __del__(self):
		self.model.stop_multi_process_pool(pool=self._pool)
		logger.debug('SentenceTransformer safely stop_multi_process_pool')

	def fit(self, X, y=0, **fit_params):
		return self

	def encode(self, sentences: Iterable[str]) -> List[np.ndarray]:
		# return self.model.encode(sentences, batch_size=self._batch_size)
		return self.model.encode_multi_process(
			[str(s) for s in sentences],
			pool=self._pool,
			batch_size=self._batch_size
		)

	@property
	def tokenizer(self) -> Tokenizer:
		return self._tokenizer

	@property
	def embedding_length(self) -> int:
		return self.model.get_sentence_embedding_dimension()


class _Tokenizer(Tokenizer):

	def __init__(self, delegator):
		self._delegator = delegator

	def tokenize(self, text: str) -> List[str]:
		return self._delegator.tokenize(text)


class AccutuningLabeler(BERTVectorizer):

	# def __init__(self, feature_name, classifier_fp, classifier_label_fp):
	def __init__(self, feature_name, classifier, classifier_labels, append_vectors=False):
		super().__init__(feature_name)

		self.classifier = classifier
		self.labels = classifier_labels
		self.append_vectors = append_vectors

	def transform(self, X, y=0):
		vector_df = self.vectorize(X)
		tags = self.classifier.predict(vector_df)
		tags = [
			self.labels[int(tag)]
			for tag in tags
		]
		tag_df = pd.DataFrame(
			tags,
			columns=[self.feature_name + '__tag']
		)
		X = X.drop(self.feature_name, axis=1)
		if self.append_vectors:
			return pd.concat(
				[X, vector_df, tag_df],
				axis=1
			)
		else:
			return pd.concat(
				[X, tag_df],
				axis=1
			)





from typing import List, Union, Iterable

import numpy as np
import pandas as pd
from flair.data import Sentence, Tokenizer
from flair.embeddings.document import DocumentEmbeddings
from sklearn.feature_extraction.text import TfidfVectorizer

from accutuning_helpers.text.embedder import TokenEmbedderBase
from accutuning_helpers.text.tokenizer import KonlpyTokenizer


class TfIdfTokenVectorizer(TokenEmbedderBase, DocumentEmbeddings):

	def __init__(
			self,
			feature_name: str,
			tokenizer: Tokenizer = KonlpyTokenizer('MeCab'),
			decode_error="replace",
			lowercase=False,
			min_df=10,  # configurable
			ngram_range=(1,3),
			max_features=500,
			**vectorizer_params,
	):
		super(TfIdfTokenVectorizer, self).__init__(feature_name)
		vec = TfidfVectorizer(
			decode_error=decode_error,
			min_df=min_df,
			tokenizer=tokenizer,
			ngram_range=ngram_range,
			max_features=max_features,
			**vectorizer_params,
		)
		self._tokenizer = tokenizer
		self._vectorizer = vec
		self.name: str = "accutuning_text_tfidf"

	def fit(self, df: pd.DataFrame, y=None) -> "TfIdfTokenVectorizer":
		texts = df[self.feature_name].tolist()
		self._vectorizer.fit(texts, y=y)
		return self

	def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
		corpus_embeddings = self.encode(X[self.feature_name].tolist())
		corpus_columns = [ f"{self.feature_name}_{v}" for v in self._vectorizer.get_feature_names()]
		X = X.drop(self.feature_name, axis=1)

		## BUGFIX df.concat(X, corpus_df, axis=1) 이 NaN을 일으켜 numpy로 concat후 X부분만 형변환 함
		concated_ = np.append(X.to_numpy(), corpus_embeddings.A, axis=1)
		concated_df = pd.DataFrame(concated_, columns=list(X.columns) + corpus_columns)
		for c in X.columns:
			concated_df[c] = concated_df[c].astype(X[c].dtype)

		return concated_df

	def encode(self, sentences: Iterable[str]) -> np.ndarray:
		# documents = [self._tokenizer.tokenize(text) for text in X]
		# return self._vectorizer.transform(documents)
		return self._vectorizer.transform(sentences)

	@property
	def tokenizer(self) -> Tokenizer:
		return self._tokenizer

	@property
	def embedding_length(self) -> int:
		vec = self._vectorizer
		return 0 if not hasattr(vec, 'vocabulary_') else len(vec.vocabulary_)

	def embed(self, sentences: Union[List[Sentence], Sentence]):
		"""Add embeddings to every sentence in the given list of sentences."""

		# if only one sentence is passed, convert to list of sentence
		if isinstance(sentences, Sentence):
			sentences = [sentences]

		import torch
		raw_sentences = [s.to_original_text() for s in sentences]
		tfidf_vectors = torch.from_numpy(self._vectorizer.transform(raw_sentences).A)

		for sentence_id, sentence in enumerate(sentences):
			sentence.set_embedding(self.name, tfidf_vectors[sentence_id])



import logging
import re
from typing import List, Tuple, Set, Union

from ckonlpy.tag import Twitter
from flair.data import Tokenizer
from konlpy.tag import Okt, Komoran, Mecab, Hannanum, Kkma

logger = logging.getLogger(__name__)

AVAILABLE_TOKENIZERS = {'okt', 'komoran', 'mecab', 'hannanum', 'kkma'}
NEG_FILTER_POS = {
	"E",  # (Verbal endings)
	"J",  # (Ending Particle)
	"JKS",  # (Josa)
	"JKB",  # (Junction)
	"JX",  # (Junction)
	"MM",  # (Modifier)
	"SP",  # (Space)
	"SSC",  # (Closing brackets)
	"SSO",  # (Opening brackets)
	"SC",  # (Separator)
	"SE",  # (Ellipsis)
	"XPN",  # (Prefix)
	"XSA",  # (Adjective Suffix)
	"XSN",  # (Noun Suffix)
	"XSV",  # (Verb Suffix)
	"UNA",  # (Unknown)
	"NA",  # (Unknown)
	"VSV",  # (Unknown)
	"VCN",  # (Negative designator)
	"VCP"  # (Positive designator)
}


# 임시로 지정


class KonlpyTokenizer(Tokenizer):
	"""
	Tokenizer using koNLP, a third party library which supports
	multiple Korean tokenizer such as MeCab, Komoran, okt, hannanum, kkma.

	default - NEG_FILTER_POS 에 속한 품사는 제외함

	For further details see:
		https://github.com/konlpy/konlpy
	"""

	def __init__(
			self,
			tokenizer_name: str,
			excluded_pos: Union[List, Set] = NEG_FILTER_POS,
	):
		super(KonlpyTokenizer, self).__init__()

		name = tokenizer_name.lower()

		if name not in AVAILABLE_TOKENIZERS:
			raise NotImplementedError(
				f"Currently, {tokenizer_name} is only supported. Supported tokenizers: {AVAILABLE_TOKENIZERS}."
			)

		if name == 'okt':
			tokenizer_ = Okt()
		elif name == 'komoran':
			tokenizer_ = Komoran()
		elif name == 'mecab':
			tokenizer_ = Mecab()
		elif name == 'hannanum':
			tokenizer_ = Hannanum()
		elif name == 'kkma':
			tokenizer_ = Kkma()

		self.tokenizer_name = tokenizer_name
		self.word_tokenizer = tokenizer_
		self._excluded_pos = excluded_pos or NEG_FILTER_POS

	def tokenize(self, text: str) -> List[str]:
		return list(map(lambda t: t[0], self.pos(text)))

	def __call__(self, text: str) -> List[str]:
		return self.tokenize(text)

	def pos(self, text: str) -> List[Tuple]:
		excluded_pos = self._excluded_pos
		pos_filtered = filter(lambda x: x[1] not in excluded_pos, self.word_tokenizer.pos(text))
		return list(pos_filtered)

	@property
	def name(self) -> str:
		return self.__class__.__name__ + "_" + self.tokenizer


VOCABS = [
	'내부', '제도', '신고서', '전자', '하이프사이클', '테스팅', '고객사', '커스터마이징', '테스터', '미연', '지구국', '메시지', '영상처리',
	'인터뷰', '조직도', '경영진'
]

DELETED_CHARS = "\\'|\\[|\\]|,|\\."


class TwitterTokenizer(Tokenizer):

	def __init__(self):
		self.pos_tagger = Twitter(use_twitter_dictionary=False)
		self.pos_tagger.add_dictionary(VOCABS, 'Noun')
		self.pos_tagger.add_a_template(('Modifier', 'Noun', 'Noun', 'Noun', 'Noun'))

	def tokenize(self, text: str) -> List[str]:
		return list(filter(lambda x: x, self.stcs_to_words(text).split(' ')))

	def __call__(self, text: str) -> List[str]:
		return self.tokenize(text)

	def stcs_to_words(self, text: str) -> str:
		stc_tagged = map(lambda t: t[0], self.pos_tagger.pos(text))
		stc = ' '.join(stc_tagged)
		stc_replaced = re.sub(DELETED_CHARS, u'', stc)
		return stc_replaced
